"""
Module to provide processing for the setext heading leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken, StackToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-few-public-methods


class SetextLeafBlockProcessor:
    """
    Class to provide processing for the setext heading leaf blocks.
    """

    __setext_characters = "-="

    @staticmethod
    def __parse_setext_headings_with_tab(
        original_line: str,
        line_to_parse: str,
        extracted_whitespace: str,
    ) -> Tuple[str, str, bool, bool, Optional[str]]:
        POGGER.debug("original_line>>:$:<", original_line)
        POGGER.debug("line_to_parse>>:$:<", line_to_parse)

        (
            token_text,
            split_tab,
            split_tab_with_block_quote_suffix,
            extra_whitespace_prefix,
            new_extracted_whitespace,
            _,
        ) = TabHelper.parse_thematic_break_with_tab(
            original_line, line_to_parse, extracted_whitespace
        )
        POGGER.debug("token_text>>:$:<", token_text)
        POGGER.debug("split_tab>>:$:<", split_tab)
        POGGER.debug(
            "split_tab_with_block_quote_suffix>>:$:<", split_tab_with_block_quote_suffix
        )
        POGGER.debug("extra_whitespace_prefix>>:$:<", extra_whitespace_prefix)
        POGGER.debug("new_extracted_whitespace>>:$:<", new_extracted_whitespace)

        return (
            token_text,
            new_extracted_whitespace,
            split_tab,
            split_tab_with_block_quote_suffix,
            extra_whitespace_prefix,
        )

    @staticmethod
    def parse_setext_headings(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        block_quote_data: BlockQuoteData,
        original_line: str,
    ) -> List[MarkdownToken]:
        """
        Handle the parsing of an setext heading.
        """

        new_tokens: List[MarkdownToken] = []
        POGGER.debug("extracted_whitespace=>:$:<", extracted_whitespace)
        if (
            TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
            and ParserHelper.is_character_at_index_one_of(
                position_marker.text_to_parse,
                position_marker.index_number,
                SetextLeafBlockProcessor.__setext_characters,
            )
            and parser_state.token_stack[-1].is_paragraph
            and (block_quote_data.current_count == block_quote_data.stack_count)
        ):
            POGGER.debug(
                "parse_setext_headings>>start",
            )
            is_paragraph_continuation = (
                SetextLeafBlockProcessor.__adjust_continuation_for_active_list(
                    parser_state, position_marker
                )
            )

            line_to_parse = position_marker.text_to_parse[
                position_marker.index_number :
            ]
            ex_ws_l = len(extracted_whitespace)
            POGGER.debug("line_to_parse=>:$:<", line_to_parse)
            POGGER.debug("original_line=>:$:<", original_line)
            split_tab = False
            split_tab_with_block_quote_suffix = False
            extra_whitespace_prefix = None
            if ParserHelper.tab_character in original_line:
                (
                    line_to_parse,
                    extracted_whitespace,
                    split_tab,
                    split_tab_with_block_quote_suffix,
                    extra_whitespace_prefix,
                ) = SetextLeafBlockProcessor.__parse_setext_headings_with_tab(
                    original_line,
                    line_to_parse,
                    extracted_whitespace,
                )

            SetextLeafBlockProcessor.__prepare_and_create_setext_token(
                parser_state,
                position_marker,
                line_to_parse,
                is_paragraph_continuation,
                split_tab,
                new_tokens,
                extracted_whitespace,
                ex_ws_l,
                split_tab_with_block_quote_suffix,
                extra_whitespace_prefix,
                block_quote_data,
            )
        else:
            POGGER.debug(
                "parse_setext_headings>>not eligible",
            )
        return new_tokens

    # pylint: disable=too-many-arguments
    @staticmethod
    def __prepare_and_create_setext_token(
        parser_state: ParserState,
        position_marker: PositionMarker,
        line_to_parse: str,
        is_paragraph_continuation: bool,
        split_tab: bool,
        new_tokens: List[MarkdownToken],
        extracted_whitespace: str,
        ex_ws_l: int,
        split_tab_with_block_quote_suffix: bool,
        extra_whitespace_prefix: Optional[str],
        block_quote_data: BlockQuoteData,
    ) -> Tuple[int, int, str]:
        _, collected_to_index = ParserHelper.collect_while_character_verified(
            line_to_parse,
            0,
            position_marker.text_to_parse[position_marker.index_number],
        )
        POGGER.debug(
            ">>position_marker.index_number>:$:<", position_marker.index_number
        )

        POGGER.debug(">>collected_to_index>:$:<", collected_to_index)
        (
            after_whitespace_index,
            extra_whitespace_after_setext,
        ) = ParserHelper.extract_spaces_verified(line_to_parse, collected_to_index)

        if not is_paragraph_continuation and after_whitespace_index == len(
            line_to_parse
        ):
            old_top_of_stack = parser_state.token_stack[-1]

            if split_tab and split_tab_with_block_quote_suffix:
                TabHelper.adjust_block_quote_indent_for_tab(parser_state)
            SetextLeafBlockProcessor.__create_setext_token(
                parser_state,
                position_marker,
                collected_to_index + ex_ws_l,
                new_tokens,
                extracted_whitespace,
                extra_whitespace_after_setext,
            )
            if split_tab and not split_tab_with_block_quote_suffix:
                SetextLeafBlockProcessor.__prepare_and_create_setext_token_list_adjust(
                    parser_state,
                    position_marker,
                    extracted_whitespace,
                    extra_whitespace_prefix,
                    old_top_of_stack,
                    new_tokens,
                    block_quote_data,
                )
        return collected_to_index, after_whitespace_index, extra_whitespace_after_setext

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __prepare_and_create_setext_token_list_adjust(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        extra_whitespace_prefix: Optional[str],
        old_top_of_stack: StackToken,
        new_tokens: List[MarkdownToken],
        block_quote_data: BlockQuoteData,
    ) -> None:
        POGGER.debug("parser_state.token_stack[-1]>>:$:<", parser_state.token_stack[-1])
        POGGER.debug("parser_state.token_stack>>:$:<", parser_state.token_stack)
        POGGER.debug("parser_state.token_document>>:$:<", parser_state.token_document)
        assert parser_state.token_stack[-1].is_list, "Should be within a list block."
        modified_whitespace = (
            extra_whitespace_prefix + extracted_whitespace
            if extra_whitespace_prefix is not None
            else extracted_whitespace
        )
        TabHelper.adjust_block_quote_indent_for_tab(parser_state, modified_whitespace)
        POGGER.debug("parser_state.token_stack>>:$:<", parser_state.token_stack)
        POGGER.debug("parser_state.token_document>>:$:<", parser_state.token_document)
        LeafBlockHelper.correct_for_leaf_block_start_in_list(
            parser_state,
            position_marker.index_indent,
            old_top_of_stack,
            new_tokens,
            block_quote_data,
            was_token_already_added_to_stack=False,
            delay_tab_match=True,
        )
        POGGER.debug("parser_state.token_stack>>:$:<", parser_state.token_stack)
        POGGER.debug("parser_state.token_document>>:$:<", parser_state.token_document)
        POGGER.debug("new_tokens>>:$:<", new_tokens)

    # pylint: enable=too-many-arguments
    @staticmethod
    def __adjust_continuation_for_active_list(
        parser_state: ParserState, position_marker: PositionMarker
    ) -> bool:
        is_paragraph_continuation: bool = (
            len(parser_state.token_stack) > 1 and parser_state.token_stack[-2].is_list
        )
        if is_paragraph_continuation:
            list_token = cast(ListStackToken, parser_state.token_stack[-2])
            POGGER.debug(
                "parser_state.original_line_to_parse>:$:<",
                parser_state.original_line_to_parse,
            )
            adj_text = position_marker.text_to_parse[position_marker.index_number :]
            assert (
                parser_state.original_line_to_parse is not None
            ), "Original line should be defined by this point."
            assert parser_state.original_line_to_parse.endswith(
                adj_text
            ), "Original line should end with the adjusted text."
            removed_text_length = len(parser_state.original_line_to_parse) - len(
                adj_text
            )
            POGGER.debug("removed_text_length>:$:<", removed_text_length)
            POGGER.debug("adj_text>:$:<", adj_text)
            POGGER.debug("indent_level>:$:<", list_token.indent_level)
            is_paragraph_continuation = (
                bool(adj_text) and removed_text_length < list_token.indent_level
            )
        return is_paragraph_continuation

    # pylint: disable=too-many-arguments
    @staticmethod
    def __create_setext_token(
        parser_state: ParserState,
        position_marker: PositionMarker,
        collected_to_index: int,
        new_tokens: List[MarkdownToken],
        extracted_whitespace: str,
        extra_whitespace_after_setext: str,
    ) -> None:
        token_index = len(parser_state.token_document) - 1
        while not parser_state.token_document[token_index].is_paragraph:
            token_index -= 1

        paragraph_token = cast(
            ParagraphMarkdownToken, parser_state.token_document[token_index]
        )
        assert (
            paragraph_token.extra_data is not None
        ), "SetExt token extra data should already be defined."
        replacement_token = SetextHeadingMarkdownToken(
            position_marker.text_to_parse[position_marker.index_number],
            collected_to_index - position_marker.index_number,
            paragraph_token.extra_data,
            position_marker,
            paragraph_token,
        )

        # This is unusual.  Normally, close_open_blocks is used to close off
        # blocks based on the stack token.  However, since the setext takes
        # the last paragraph of text (see case 61) and translates it
        # into a heading, this has to be done separately, as there is no
        # stack token to close.
        new_tokens.append(
            replacement_token.generate_close_markdown_token_from_markdown_token(
                extracted_whitespace, extra_whitespace_after_setext
            )
        )

        parser_state.token_document[token_index] = replacement_token
        del parser_state.token_stack[-1]

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to orchestrate the handling of the different inline elements.
"""

import logging
from typing import Dict, List, Optional, Set, Tuple, cast

from typing_extensions import Protocol

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.extension_manager.extension_manager import ExtensionManager
from pymarkdown.extensions.extended_autolinks import MarkdownExtendedAutolinksExtension
from pymarkdown.extensions.task_list_items import TaskListToken
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.emphasis_helper import EmphasisHelper
from pymarkdown.inline.inline_autolink_helper import InlineAutoLinkHelper
from pymarkdown.inline.inline_backslash_helper import InlineBackslashHelper
from pymarkdown.inline.inline_backtick_helper import InlineBacktickHelper
from pymarkdown.inline.inline_character_reference_helper import (
    InlineCharacterReferenceHelper,
)
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse
from pymarkdown.links.link_parse_helper import LinkParseHelper
from pymarkdown.links.link_search_helper import LinkSearchHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.reference_markdown_token import ReferenceMarkdownToken
from pymarkdown.tokens.special_text_markdown_token import SpecialTextMarkdownToken

# pylint: disable=too-many-lines

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods
class InlineHandlerProtocol(Protocol):
    """
    Protocol to handle the various inline sequences.
    """

    def __call__(  # noqa: E704
        self,
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
    ) -> InlineResponse: ...  # pragma: no cover


# pylint: enable=too-few-public-methods


class InlineHandlerHelper:
    """
    Class to orchestrate the handling of the different inline elements.
    """

    valid_inline_text_block_sequence_starts = ""
    __valid_inline_simple_text_block_sequence_starts = ""

    __inline_character_handlers: Dict[str, InlineHandlerProtocol] = {}
    __inline_simple_character_handlers: Dict[str, InlineHandlerProtocol] = {}

    __inline_processing_needed = ""

    @staticmethod
    def initialize(extension_manager: ExtensionManager) -> None:
        """
        Initialize the inline processor subsystem.
        """
        EmphasisHelper.initialize(extension_manager)
        InlineHandlerHelper.__inline_processing_needed = (
            f"{LinkParseHelper.link_label_start}{LinkParseHelper.link_label_end}"
        )
        InlineHandlerHelper.__inline_processing_needed += (
            EmphasisHelper.get_inline_emphasis()
        )

        InlineHandlerHelper.__inline_character_handlers = {}
        InlineHandlerHelper.__inline_simple_character_handlers = {}
        InlineHandlerHelper.valid_inline_text_block_sequence_starts = (
            ParserHelper.newline_character
        )
        InlineHandlerHelper.__valid_inline_simple_text_block_sequence_starts = (
            ParserHelper.newline_character
        )

        InlineHandlerHelper.register_handlers(
            InlineBacktickHelper.code_span_bounds,
            InlineBacktickHelper.handle_inline_backtick,
        )
        InlineHandlerHelper.register_handlers(
            InlineBackslashHelper.backslash_character,
            InlineBackslashHelper.handle_inline_backslash,
            is_simple_handler=True,
        )
        InlineHandlerHelper.register_handlers(
            InlineCharacterReferenceHelper.character_reference_start_character,
            InlineCharacterReferenceHelper.handle_character_reference,
            is_simple_handler=True,
        )
        InlineHandlerHelper.register_handlers(
            InlineAutoLinkHelper.angle_bracket_start,
            InlineAutoLinkHelper.handle_angle_brackets,
        )
        for i in InlineHandlerHelper.__inline_processing_needed:
            InlineHandlerHelper.register_handlers(
                i, InlineHandlerHelper.__handle_inline_special_single_character
            )
        InlineHandlerHelper.register_handlers(
            LinkSearchHelper.image_start_sequence[0],
            InlineHandlerHelper.__handle_inline_image_link_start_character,
        )
        for i in ParserHelper.valid_characters_to_escape():
            InlineHandlerHelper.register_handlers(
                i, InlineHandlerHelper.__handle_inline_control_character
            )
        if extension_manager.is_extended_autolinks_enabled:
            InlineHandlerHelper.register_handlers(
                "h",
                MarkdownExtendedAutolinksExtension.handle_http_autolink,
            )
            InlineHandlerHelper.register_handlers(
                "w",
                MarkdownExtendedAutolinksExtension.handle_www_autolink,
            )
            InlineHandlerHelper.register_handlers(
                "@",
                MarkdownExtendedAutolinksExtension.handle_email_autolink,
            )
            InlineHandlerHelper.register_handlers(
                "x",
                MarkdownExtendedAutolinksExtension.handle_email_autolink,
            )
            InlineHandlerHelper.register_handlers(
                "m",
                MarkdownExtendedAutolinksExtension.handle_email_autolink,
            )

    @staticmethod
    def __handle_inline_control_character(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
    ) -> InlineResponse:
        """
        Registered handler.
        """
        _ = parser_properties

        inline_response = InlineResponse()
        (
            inline_response.new_index,
            inline_response.new_string,
            inline_response.delta_column_number,
        ) = (
            inline_request.next_index + 1,
            f"{ParserHelper.escape_character}{inline_request.source_text[inline_request.next_index]}",
            1,
        )
        return inline_response

    @staticmethod
    def has_handler(inline_character: str) -> bool:
        """
        Check to see if a handler is registered.
        """
        return inline_character in InlineHandlerHelper.__inline_character_handlers

    @staticmethod
    def __get_handler(
        inline_character: str,
    ) -> InlineHandlerProtocol:
        return InlineHandlerHelper.__inline_character_handlers[inline_character]

    @staticmethod
    def register_handlers(
        inline_character: str,
        start_token_handler: InlineHandlerProtocol,
        is_simple_handler: bool = False,
    ) -> None:
        """
        Register the handlers necessary to deal with token's start and end.
        """
        InlineHandlerHelper.__inline_character_handlers[inline_character] = (
            start_token_handler
        )
        InlineHandlerHelper.valid_inline_text_block_sequence_starts += inline_character
        if is_simple_handler:
            InlineHandlerHelper.__inline_simple_character_handlers[inline_character] = (
                start_token_handler
            )
            InlineHandlerHelper.__valid_inline_simple_text_block_sequence_starts += (
                inline_character
            )

    @staticmethod
    def process_simple_inline_fn(
        parser_properties: ParseBlockPassProperties, source_text: str
    ) -> str:
        """
        Handle a simple processing of inline text for simple replacements.
        """

        start_index, processed_parts = 0, []
        next_index = ParserHelper.index_any_of(
            source_text,
            InlineHandlerHelper.__valid_inline_simple_text_block_sequence_starts,
            start_index,
        )
        while next_index != -1:
            processed_parts.append(source_text[start_index:next_index])
            inline_request = InlineRequest(source_text, next_index)
            if (
                source_text[next_index]
                in InlineHandlerHelper.__inline_character_handlers
            ):
                proc_fn = InlineHandlerHelper.__inline_character_handlers[
                    source_text[next_index]
                ]
                # assert proc_fn is not None
                inline_response = proc_fn(parser_properties, inline_request)
                assert (
                    inline_response.new_string is not None
                ), "new_string should be defined by this point."
                assert (
                    inline_response.new_index is not None
                ), "new_index should be defined by this point."
                processed_parts.append(inline_response.new_string)
                start_index = inline_response.new_index
            else:
                processed_parts.append(ParserHelper.newline_character)
                start_index = next_index + 1
            next_index = ParserHelper.index_any_of(
                source_text,
                InlineHandlerHelper.__valid_inline_simple_text_block_sequence_starts,
                start_index,
            )

        processed_parts.append(source_text[start_index:])
        return "".join(processed_parts)

    @staticmethod
    def __handle_inline_special_single_character(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
    ) -> InlineResponse:
        """
        Registered handler.
        """
        return InlineHandlerHelper.__handle_inline_special(
            parser_properties,
            inline_request,
            1,
            inline_request.last_container_token,
        )

    @staticmethod
    def __handle_inline_special(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
        special_length: int,
        last_container_token: Optional[MarkdownToken],
    ) -> InlineResponse:
        """
        Handle the collection of special inline characters for later processing.
        """
        inline_response = InlineResponse()
        inline_response.new_string = ""

        assert (
            inline_request.line_number is not None
            and inline_request.column_number is not None
            and inline_request.remaining_line is not None
        ), "These fields must be defined by now."

        POGGER.debug(">>tabified_text>:$:<", inline_request.tabified_text)

        POGGER.debug(">>remaining_line>>$<<", inline_request.remaining_line)
        POGGER.debug(">>column_number>>$<<", inline_request.column_number)
        column_number = inline_request.column_number + len(
            inline_request.remaining_line
        )
        POGGER.debug(">>column_number>>$<<", column_number)

        (
            special_sequence,
            inline_response.delta_column_number,
            inline_response.new_index,
            surrounding_whitespace_pair,
            is_active,
            inline_response.new_tokens,
            inline_response.consume_rest_of_line,
            inline_response.delta_line_number,
        ) = InlineHandlerHelper.__handle_inline_special_character(
            parser_properties,
            inline_request,
            special_length,
            last_container_token,
        )
        if not inline_response.new_tokens:
            POGGER.debug(">>create>>$,$<<", inline_request.line_number, column_number)
            inline_response.new_tokens = [
                SpecialTextMarkdownToken(
                    special_sequence,
                    inline_response.delta_column_number,
                    surrounding_whitespace_pair[0],
                    surrounding_whitespace_pair[1],
                    is_active,
                    inline_request.line_number,
                    column_number,
                )
            ]

        POGGER.debug(">>delta_line>>$<<", inline_response.delta_line_number)
        POGGER.debug(">>repeat_count>>$<<", inline_response.delta_column_number)
        return inline_response

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_inline_special_character_task_list(
        parser_properties: ParseBlockPassProperties,
        para_owner: Optional[ParagraphMarkdownToken],
        special_sequence: str,
        next_index: int,
        source_text: str,
        last_container_token: Optional[MarkdownToken],
    ) -> Tuple[int, Optional[MarkdownToken]]:
        repeat_count = 0
        new_token = None
        if para_owner is None or not parser_properties.is_task_lists_enabled:
            return repeat_count, new_token
        if not (last_container_token and last_container_token.is_list_start):
            return repeat_count, new_token
        if last_container_token.line_number != para_owner.line_number:
            last_list_token = cast(ListStartMarkdownToken, last_container_token)
            if (
                last_list_token.last_new_list_token is None
                or last_list_token.last_new_list_token.line_number
                != para_owner.line_number
            ):
                return repeat_count, new_token

        if special_sequence == "[" and next_index == 0 and len(source_text) >= 4:
            is_valid_check_character = source_text[1] in [" ", "X", "x"]
            is_valid_close_character = source_text[2] == "]"
            is_followed_by_whitespace = ParserHelper.is_character_at_index_whitespace(
                source_text, 3
            )
            if (
                is_valid_check_character
                and is_valid_close_character
                and is_followed_by_whitespace
            ):
                repeat_count = 3
                new_token = TaskListToken(
                    source_text[1],
                    line_number=para_owner.line_number,
                    column_number=para_owner.column_number,
                )
        return repeat_count, new_token

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_inline_special_character(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
        special_length: int,
        last_container_token: Optional[MarkdownToken],
    ) -> Tuple[
        str,
        int,
        int,
        Tuple[Optional[str], Optional[str]],
        bool,
        List[MarkdownToken],
        bool,
        int,
    ]:
        assert (
            inline_request.current_string_unresolved is not None
        ), "These fields must be defined by now."

        special_sequence = inline_request.source_text[
            inline_request.next_index : inline_request.next_index + special_length
        ]
        if special_length == 1:
            (
                repeat_count,
                new_token,
            ) = InlineHandlerHelper.__handle_inline_special_character_task_list(
                parser_properties,
                inline_request.para_owner,
                special_sequence,
                inline_request.next_index,
                inline_request.source_text,
                last_container_token,
            )
            new_index = inline_request.next_index + repeat_count
            if new_token is not None:
                return (
                    inline_request.source_text[inline_request.next_index : new_index],
                    repeat_count,
                    new_index,
                    (None, None),
                    True,
                    [new_token],
                    False,
                    0,
                )

            if special_sequence in EmphasisHelper.get_inline_emphasis():
                return InlineHandlerHelper.__handle_inline_special_character_emphasis(
                    inline_request.source_text,
                    inline_request.next_index,
                    special_sequence,
                )
        if special_sequence[0] == LinkParseHelper.link_label_end:
            POGGER.debug(
                "POSSIBLE LINK CLOSE_FOUND($)>>$>>",
                special_length,
                special_sequence,
            )

            return InlineHandlerHelper.__handle_link_label_end(
                parser_properties,
                inline_request,
                special_sequence,
                inline_request.current_string_unresolved,
            )
        return (
            special_sequence,
            special_length,
            inline_request.next_index + special_length,
            (None, None),
            True,
            [],
            False,
            0,
        )

    @staticmethod
    def __handle_inline_special_character_emphasis(
        source_text: str,
        next_index: int,
        special_sequence: str,
    ) -> Tuple[
        str,
        int,
        int,
        Tuple[Optional[str], Optional[str]],
        bool,
        List[MarkdownToken],
        bool,
        int,
    ]:
        repeat_count, new_index = ParserHelper.collect_while_character_verified(
            source_text, next_index, special_sequence
        )
        return (
            source_text[next_index:new_index],
            repeat_count,
            new_index,
            (
                source_text[max(0, next_index - 2) : next_index],
                source_text[new_index : min(len(source_text), new_index + 2)],
            ),
            True,
            [],
            False,
            0,
        )

    @staticmethod
    def __handle_inline_image_link_start_character(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
    ) -> InlineResponse:
        """
        Registered handler.
        """
        if ParserHelper.are_characters_at_index(
            inline_request.source_text,
            inline_request.next_index,
            LinkSearchHelper.image_start_sequence,
        ):
            assert (
                inline_request.line_number is not None
                and inline_request.column_number is not None
                and inline_request.remaining_line is not None
                and inline_request.current_string_unresolved is not None
            ), "These four fields must be defined by now."

            inline_response = InlineHandlerHelper.__handle_inline_special(
                parser_properties,
                inline_request,
                2,
                None,
            )
            assert (
                not inline_response.consume_rest_of_line
            ), "After the processing, make sure we are not just consuming the rest of the line."
        else:
            inline_response = InlineResponse()
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.delta_column_number,
            ) = (
                LinkSearchHelper.image_start_sequence[0],
                inline_request.next_index + 1,
                1,
            )
        return inline_response

    @staticmethod
    def __handle_link_label_end(
        parser_properties: ParseBlockPassProperties,
        inline_request: InlineRequest,
        special_sequence: str,
        current_string_unresolved: str,
    ) -> Tuple[
        str,
        int,
        int,
        Tuple[Optional[str], Optional[str]],
        bool,
        List[MarkdownToken],
        bool,
        int,
    ]:
        assert (
            inline_request.remaining_line is not None
        ), "These fields must be defined by now."
        inline_blocks = inline_request.inline_blocks

        POGGER.debug(
            ">>inline_blocks>>$<<",
            inline_blocks,
        )
        POGGER.debug(
            ">>remaining_line>>$<<",
            inline_request.remaining_line,
        )
        POGGER.debug(
            ">>tabified_remaining_line>>$<<",
            inline_request.tabified_remaining_line,
        )
        POGGER.debug(
            ">>current_string_unresolved>>$<<",
            current_string_unresolved,
        )
        POGGER.debug(
            ">>source_text[next_index=$:]>:$:<",
            inline_request.next_index,
            inline_request.source_text[inline_request.next_index :],
        )
        POGGER.debug(
            ">>source_text>:$:<",
            inline_request.source_text,
        )
        POGGER.debug(
            ">>tabified_text>>$<<",
            inline_request.tabified_text,
        )
        POGGER.debug("")
        old_inline_blocks_count, old_inline_blocks_last_token = (
            len(inline_blocks),
            inline_blocks[-1] if inline_blocks else None,
        )

        (
            new_index,
            is_active,
            new_token,
            consume_rest_of_line,
        ) = LinkSearchHelper.look_for_link_or_image(
            parser_properties,
            inline_blocks,
            inline_request.source_text,
            inline_request.next_index,
            inline_request.remaining_line,
            inline_request.tabified_remaining_line,
            current_string_unresolved,
            InlineHandlerHelper.process_simple_inline_fn,
            inline_request.tabified_text,
        )
        POGGER.debug(">>next_index>>$<<", inline_request.next_index)
        POGGER.debug(">>new_index>>$<<", new_index)
        POGGER.debug(
            ">>source_text:new_index>>$<<",
            inline_request.source_text[new_index:],
        )
        POGGER.debug(">>inline_blocks>>$<<", inline_blocks)
        POGGER.debug(">>new_token>>$<<", new_token)
        POGGER.debug(">>source_text>>$<<", inline_request.source_text[new_index:])
        POGGER.debug(">>consume_rest_of_line>>$<<", consume_rest_of_line)
        POGGER.debug(">>old_inline_blocks_count>>$<<", old_inline_blocks_count)

        return InlineHandlerHelper.__handle_link_label_end_calc(
            0,
            new_token,
            inline_blocks,
            new_index,
            inline_request.next_index,
            len(inline_request.remaining_line),
            inline_request.para_owner,
            old_inline_blocks_count,
            old_inline_blocks_last_token,
            is_active,
            consume_rest_of_line,
            special_sequence,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_link_label_end_calc(
        delta_line: int,
        new_token: Optional[MarkdownToken],
        inline_blocks: List[MarkdownToken],
        new_index: int,
        next_index: int,
        remaining_line_size: int,
        para_owner: Optional[ParagraphMarkdownToken],
        old_inline_blocks_count: int,
        old_inline_blocks_last_token: Optional[MarkdownToken],
        is_active: bool,
        consume_rest_of_line: bool,
        special_sequence: str,
    ) -> Tuple[
        str,
        int,
        int,
        Tuple[Optional[str], Optional[str]],
        bool,
        List[MarkdownToken],
        bool,
        int,
    ]:
        repeat_count = 1
        new_inline_blocks_count = len(inline_blocks)
        POGGER.debug(">>new_inline_blocks_count>>$<<", new_inline_blocks_count)
        if (
            new_token
            or old_inline_blocks_count != new_inline_blocks_count
            or (inline_blocks and old_inline_blocks_last_token != inline_blocks[-1])
        ):
            (
                delta_line,
                repeat_count,
            ) = InlineHandlerHelper.__calculate_repeat_count_and_delta_line(
                inline_blocks,
                new_token,
                new_index,
                next_index,
                remaining_line_size,
                para_owner,
                delta_line,
            )
        return (
            special_sequence,
            repeat_count,
            new_index,
            (None, None),
            is_active,
            [new_token] if new_token else [],
            consume_rest_of_line,
            delta_line,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def process_inline_handled_character(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        next_index: int,
        inline_request: InlineRequest,
        line_number: int,
        column_number: int,
        coalesced_stack: List[MarkdownToken],
        pragma_line_numbers: Set[int],
    ) -> Tuple[InlineResponse, int, int, bool, bool]:
        """
        Process the handler specified by the next character.
        """

        proc_fn = InlineHandlerHelper.__get_handler(source_text[next_index])
        # assert proc_fn is not None
        inline_response = proc_fn(parser_properties, inline_request)

        for _ in range(inline_response.delta_line_number):
            line_number += 1
            while line_number in pragma_line_numbers:
                line_number += 1

        did_line_number_change = bool(inline_response.delta_line_number)
        was_column_number_reset = inline_response.delta_column_number < 0
        column_number = (
            -inline_response.delta_column_number
            if was_column_number_reset
            else column_number + inline_response.delta_column_number
        )

        if (
            coalesced_stack
            and coalesced_stack[-1].is_block_quote_start
            and (
                inline_response.new_tokens
                and inline_response.new_tokens[-1].is_inline_raw_html
            )
        ):
            raw_html_token = cast(RawHtmlMarkdownToken, inline_response.new_tokens[-1])
            newline_count = ParserHelper.count_newlines_in_text(raw_html_token.raw_tag)
            POGGER.debug("newline_count in raw-html>>$>", newline_count)

            block_quote_token = cast(BlockQuoteMarkdownToken, coalesced_stack[-1])
            block_quote_token.leading_text_index += newline_count

        return (
            inline_response,
            line_number,
            column_number,
            was_column_number_reset,
            did_line_number_change,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_repeat_count_and_delta_line(
        inline_blocks: List[MarkdownToken],
        new_token: Optional[MarkdownToken],
        new_index: int,
        next_index: int,
        remaining_line_size: int,
        para_owner: Optional[ParagraphMarkdownToken],
        delta_line: int,
    ) -> Tuple[int, int]:
        if inline_blocks[-1].is_inline_image:
            repeat_count = (new_index - next_index) + remaining_line_size
            (
                delta_line,
                repeat_count,
            ) = InlineHandlerHelper.__calculate_link_and_image_deltas(
                para_owner, inline_blocks[-1], delta_line, repeat_count
            )
            POGGER.debug(">>delta_line>>$<<", delta_line)
            POGGER.debug(">>repeat_count>>$<<", repeat_count)
        elif new_token and new_token.is_inline_link_end:
            end_token = cast(EndMarkdownToken, new_token)
            POGGER.debug(
                ">>new_token.start_markdown_token>>$<<",
                end_token.start_markdown_token,
            )
            assert (
                end_token.start_markdown_token
            ), "End markdown tokens should have the starting token."
            repeat_count = new_index - next_index
            POGGER.debug(">>delta_line>>$<<", delta_line)
            POGGER.debug(">>repeat_count>>$<<", repeat_count)
            if para_owner:
                POGGER.debug(
                    ">>para_owner.rehydrate_index>>$<<",
                    para_owner.rehydrate_index,
                )
            (
                delta_line,
                repeat_count,
            ) = InlineHandlerHelper.__calculate_link_and_image_deltas(
                para_owner,
                end_token.start_markdown_token,
                delta_line,
                repeat_count,
            )
            if para_owner:
                POGGER.debug(
                    ">>para_owner.rehydrate_index>>$<<",
                    para_owner.rehydrate_index,
                )
            POGGER.debug(">>delta_line>>$<<", delta_line)
            POGGER.debug(">>repeat_count>>$<<", repeat_count)
        else:
            repeat_count = new_index - next_index
            POGGER.debug(">>repeat_count>>$<<", repeat_count)
        return delta_line, repeat_count

    # pylint: enable=too-many-arguments

    @staticmethod
    def __calculate_full_deltas(
        current_token: ReferenceMarkdownToken,
        para_owner: Optional[ParagraphMarkdownToken],
        delta_line: int,
        repeat_count: int,
    ) -> Tuple[int, int]:
        assert (
            current_token.ex_label is not None
        ), "All reference tokens must have a label."
        if newline_count := ParserHelper.count_newlines_in_text(current_token.ex_label):
            POGGER.debug(">>ex_label")
            delta_line += newline_count
            if para_owner:
                POGGER.debug(
                    ">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index
                )
                para_owner.rehydrate_index += newline_count
                POGGER.debug(
                    ">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index
                )
            POGGER.debug("full>>ex_label>>newline_count>>$", newline_count)

            last_line_of_label = ParserHelper.calculate_last_line(
                current_token.ex_label
            )
            repeat_count = -(len(last_line_of_label) + 2)
        return delta_line, repeat_count

    @staticmethod
    def __calculate_inline_reference_deltas(
        current_token: ReferenceMarkdownToken,
        para_owner: Optional[ParagraphMarkdownToken],
        split_paragraph_lines: Optional[List[str]],
        delta_line: int,
        repeat_count: int,
    ) -> Tuple[int, int]:
        active_link_title = current_token.active_link_title

        assert (
            current_token.before_title_whitespace is not None
        ), "For an inline, this must be defined."
        assert (
            current_token.before_link_whitespace is not None
        ), "For an inline, this must be defined."
        link_part_lengths = [0] * 5
        link_part_lengths[0] = len(current_token.active_link_uri) + len(
            current_token.before_title_whitespace
        )
        if current_token.inline_title_bounding_character:
            assert (
                active_link_title is not None
                and current_token.after_title_whitespace is not None
            ), "If we have an inline link, both these should be defined."

            link_part_lengths[1] = 1
            link_part_lengths[2] = len(active_link_title) + 1
            link_part_lengths[3] = len(current_token.after_title_whitespace)
        POGGER.debug(">>link_part_lengths>>$<<", link_part_lengths)

        (
            link_part_index,
            total_newlines,
        ) = InlineHandlerHelper.__calculate_inline_label(current_token)

        (
            link_part_index,
            delta_line,
            last_spaces,
        ) = InlineHandlerHelper.__calculate_inline_whitespace(
            current_token.before_link_whitespace,
            "before_link_whitespace",
            0,
            link_part_index,
            delta_line,
            "",
        )

        (
            link_part_index,
            delta_line,
            last_spaces,
        ) = InlineHandlerHelper.__calculate_inline_whitespace(
            current_token.before_title_whitespace,
            "before_title_whitespace",
            1,
            link_part_index,
            delta_line,
            last_spaces,
        )

        assert active_link_title is not None, "For an inline, this must be defined."
        (
            link_part_index,
            delta_line,
            last_spaces,
            new_link_part_length,
        ) = InlineHandlerHelper.__calculate_inline_link_title(
            active_link_title, link_part_index, delta_line, last_spaces
        )
        if new_link_part_length is not None:
            link_part_lengths[2] = new_link_part_length

        assert (
            current_token.after_title_whitespace is not None
        ), "For an inline, this must be defined."
        (
            link_part_index,
            delta_line,
            last_spaces,
        ) = InlineHandlerHelper.__calculate_inline_whitespace(
            current_token.after_title_whitespace,
            "after_title_whitespace",
            4,
            link_part_index,
            delta_line,
            last_spaces,
        )

        return InlineHandlerHelper.__calculate_inline_delta_adjustments(
            link_part_index,
            total_newlines,
            delta_line,
            repeat_count,
            para_owner,
            split_paragraph_lines,
            link_part_lengths,
            last_spaces,
        )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_inline_delta_adjustments(
        link_part_index: int,
        total_newlines: int,
        delta_line: int,
        repeat_count: int,
        para_owner: Optional[ParagraphMarkdownToken],
        split_paragraph_lines: Optional[List[str]],
        link_part_lengths: List[int],
        last_spaces: str,
    ) -> Tuple[int, int]:
        POGGER.debug(">>link_part_index>>$<<", link_part_index)
        POGGER.debug(">>total_newlines>>$<<", total_newlines)
        POGGER.debug(">>delta_line>>$<<", delta_line)

        if para_owner:
            POGGER.debug(
                ">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index
            )
            para_owner.rehydrate_index += delta_line
            POGGER.debug(
                ">>para_owner.rehydrate_index>>$<<", para_owner.rehydrate_index
            )
        if link_part_index >= 0:
            if split_paragraph_lines:
                assert (
                    para_owner is not None
                ), "If we have a paragraph, must have a paragraph token."
                link_part_lengths[4] = len(
                    split_paragraph_lines[para_owner.rehydrate_index]
                )
            else:
                link_part_lengths[4] = len(
                    ParserHelper.calculate_last_line(last_spaces)
                )
            link_part_lengths[:link_part_index] = [0] * link_part_index
            repeat_count = -(2 + sum(link_part_lengths))
            POGGER.debug(">>link_part_lengths>>$<<", link_part_lengths)
            POGGER.debug(">>repeat_count>>$<<", delta_line)
        return delta_line, repeat_count

    # pylint: enable=too-many-arguments

    @staticmethod
    def __calculate_inline_label(current_token: MarkdownToken) -> Tuple[int, int]:
        reference_token = cast(ReferenceMarkdownToken, current_token)
        total_newlines = ParserHelper.count_newlines_in_text(
            reference_token.text_from_blocks
        )
        link_part_index = -1 if total_newlines else -2
        return link_part_index, total_newlines

    # pylint: disable=too-many-arguments
    @staticmethod
    def __calculate_inline_whitespace(
        sample_string: str,
        sample_name: str,
        new_link_part_index: int,
        link_part_index: int,
        delta_line: int,
        last_spaces: str,
    ) -> Tuple[int, int, str]:
        if newline_count := ParserHelper.count_newlines_in_text(sample_string):
            POGGER.debug(">>$", sample_name)
            link_part_index, delta_line, last_spaces = (
                new_link_part_index,
                delta_line + newline_count,
                sample_string[:],
            )
        return link_part_index, delta_line, last_spaces

    # pylint: enable=too-many-arguments

    @staticmethod
    def __calculate_inline_link_title(
        active_link_title: str, link_part_index: int, delta_line: int, last_spaces: str
    ) -> Tuple[int, int, str, Optional[int]]:
        if newline_count := ParserHelper.count_newlines_in_text(active_link_title):
            POGGER.debug(">>active_link_title")
            _, delta_column_number = ParserHelper.calculate_deltas(active_link_title)
            link_part_index, delta_line, last_spaces, new_link_part_length = (
                2,
                delta_line + newline_count,
                "",
                -delta_column_number,
            )
        else:
            new_link_part_length = None
        return link_part_index, delta_line, last_spaces, new_link_part_length

    @staticmethod
    def __calculate_shortcut_collapsed_deltas(
        current_token: ReferenceMarkdownToken, delta_line: int, repeat_count: int
    ) -> Tuple[int, int]:
        """
        Tests test_reference_links_extra_03jx and test_reference_links_extra_03ja added
        to ensure that this is correct.  Those tests confirm that any newlines in the
        label are already accounted for, and as such, do not require any further
        modifications.
        """
        _ = current_token
        return delta_line, repeat_count

    @staticmethod
    def __calculate_link_and_image_deltas(
        para_owner: Optional[ParagraphMarkdownToken],
        current_token: MarkdownToken,
        delta_line: int,
        repeat_count: int,
    ) -> Tuple[int, int]:
        POGGER.debug(">>delta_line>>$<<", delta_line)
        POGGER.debug(">>repeat_count>>$<<", repeat_count)

        if ParserHelper.newline_character in str(current_token):
            POGGER.debug(">>para_owner>>$<<", para_owner)
            split_paragraph_lines: Optional[List[str]] = (
                para_owner.extracted_whitespace.split(ParserHelper.newline_character)
                if para_owner
                else None
            )

            reference_token = cast(ReferenceMarkdownToken, current_token)
            POGGER.debug(">>current_token.label_type>>$<<", reference_token.label_type)
            if reference_token.label_type == Constants.link_type__inline:
                (
                    delta_line,
                    repeat_count,
                ) = InlineHandlerHelper.__calculate_inline_reference_deltas(
                    reference_token,
                    para_owner,
                    split_paragraph_lines,
                    delta_line,
                    repeat_count,
                )
            elif reference_token.label_type == Constants.link_type__full:
                delta_line, repeat_count = InlineHandlerHelper.__calculate_full_deltas(
                    reference_token, para_owner, delta_line, repeat_count
                )
            else:
                assert reference_token.label_type in (
                    Constants.link_type__shortcut,
                    Constants.link_type__collapsed,
                ), f"Label type '{reference_token.label_type}' not handled."
                (
                    delta_line,
                    repeat_count,
                ) = InlineHandlerHelper.__calculate_shortcut_collapsed_deltas(
                    reference_token, delta_line, repeat_count
                )

        POGGER.debug(">>delta_line>>$<<repeat_count>>$<<", delta_line, repeat_count)
        return delta_line, repeat_count



from typing import List, Optional, Tuple

from pymarkdown.container_blocks.container_grab_bag import POGGER
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.html.html_helper import HtmlHelper
from pymarkdown.leaf_blocks.atx_leaf_block_processor import AtxLeafBlockProcessor
from pymarkdown.leaf_blocks.fenced_leaf_block_processor import FencedLeafBlockProcessor
from pymarkdown.leaf_blocks.table_block_tuple import TableColumn, TableRow, TableTuple
from pymarkdown.leaf_blocks.thematic_leaf_block_processor import (
    ThematicLeafBlockProcessor,
)
from pymarkdown.links.link_reference_definition_parse_helper import (
    LinkReferenceDefinitionParseHelper,
)


## Different from LRD.
class TableParseHelper:
    __table_column_separator_character = "|"

    @staticmethod
    def parse_table(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        is_blank_line: bool,
        remaining_line_to_parse: Optional[str],
        was_started: bool,
    ) -> Tuple[bool, int, Optional[TableTuple]]:

        if was_started:
            if remaining_line_to_parse is not None:
                did_start = TableParseHelper.__parse_table_1(
                    parser_state, remaining_line_to_parse
                )
            else:
                did_start = True
        else:
            did_start = TableParseHelper.__is_table_start(
                parser_state, line_to_parse, start_index, extracted_whitespace
            )
        if not did_start:
            POGGER.debug("BAIL")
            return False, -1, None

        column_separator_characters = (
            TableParseHelper.__table_column_separator_character + "\n"
        )
        # multiple lines

        xyz: List[TableRow] = []
        col_as: List[Optional[str]] = []

        while start_index < len(line_to_parse):

            new_index, start_index, extracted_whitespace = (
                TableParseHelper.__parse_table_4(
                    line_to_parse,
                    start_index,
                    column_separator_characters,
                    xyz,
                    extracted_whitespace,
                )
            )

            if len(xyz) == 2:
                did_parse_separators = TableParseHelper.__parse_table_2(xyz, col_as)
                if not did_parse_separators:
                    return False, -1, None

        keep_going = is_blank_line and len(xyz) >= 2
        if keep_going:
            return TableParseHelper.__create_table_token(new_index, xyz, col_as)
        else:
            return keep_going, new_index, None

    @staticmethod
    def __parse_table_1(
        parser_state: ParserState, remaining_line_to_parse: str
    ) -> bool:
        remaining_line_start_index, remaining_line_extracted_whitespace = (
            ParserHelper.extract_spaces_verified(remaining_line_to_parse, 0)
        )

        is_atx, _, _, _ = AtxLeafBlockProcessor.is_atx_heading(
            remaining_line_to_parse,
            remaining_line_start_index,
            remaining_line_extracted_whitespace,
        )
        is_leaf_block_start = is_atx
        if not is_leaf_block_start:
            is_thematic_break_start, _ = ThematicLeafBlockProcessor.is_thematic_break(
                remaining_line_to_parse,
                remaining_line_start_index,
                remaining_line_extracted_whitespace,
            )
            is_leaf_block_start = bool(is_thematic_break_start)
        if not is_leaf_block_start:
            # TODO may have to adjust start_index for tabs
            is_indented_code_block_start = TabHelper.is_length_greater_than_or_equal_to(
                remaining_line_extracted_whitespace, 4, start_index=0
            )
            is_leaf_block_start = is_indented_code_block_start
        if not is_leaf_block_start:
            is_fenced_start, _, _, _, _ = FencedLeafBlockProcessor.is_fenced_code_block(
                parser_state,
                remaining_line_to_parse,
                remaining_line_start_index,
                remaining_line_extracted_whitespace,
                "",  # TODO may have issues with this
                0,
            )
            is_leaf_block_start = is_fenced_start
        if not is_leaf_block_start:
            is_html_start, _ = HtmlHelper.is_html_block(
                remaining_line_to_parse,
                remaining_line_start_index,
                remaining_line_extracted_whitespace,
                parser_state.token_stack,
                parser_state.parse_properties,
            )
            is_leaf_block_start = bool(is_html_start)
        if not is_leaf_block_start:
            is_lrd_start = (
                LinkReferenceDefinitionParseHelper.is_link_reference_definition(
                    parser_state,
                    remaining_line_to_parse,
                    remaining_line_start_index,
                    remaining_line_extracted_whitespace,
                )
            )
            is_leaf_block_start = is_lrd_start

        return not is_leaf_block_start

    @staticmethod
    def __parse_table_4(
        line_to_parse: str,
        start_index: int,
        column_separator_characters: str,
        xyz: List[TableRow],
        extracted_whitespace: str,
    ) -> Tuple[int, int, str]:
        did_start_with_separator = False
        if ParserHelper.is_character_at_index(
            line_to_parse,
            start_index,
            TableParseHelper.__table_column_separator_character,
        ):
            new_index = start_index + 1
            did_start_with_separator = True
        else:
            new_index = start_index

        tr_columns, trailing_whitespace, pre_h, new_index = (
            TableParseHelper.__parse_table_3(
                line_to_parse, column_separator_characters, new_index
            )
        )

        tr = TableRow(
            extracted_whitespace,
            trailing_whitespace,
            columns=tr_columns,
            end_of_row=None,
            did_start_with_separator=did_start_with_separator,
        )
        xyz.append(tr)

        start_index = new_index

        (start_index, extracted_whitespace) = ParserHelper.extract_spaces_verified(
            line_to_parse,
            new_index,
        )
        return new_index, start_index, extracted_whitespace

    @staticmethod
    def __parse_table_3a(
        tr_columns: List[TableColumn],
        extracted_sep_text: str,
        trailing_whitespace: str,
        previous_extracted_sep_text: str,
        line_to_parse: str,
    ) -> Tuple[str, int]:
        first_non_matching_index, _ = ParserHelper.collect_while_spaces(
            extracted_sep_text, 0
        )
        if first_non_matching_index == len(extracted_sep_text):
            trailing_whitespace = extracted_sep_text
        else:
            tr_columns.append(
                TableColumn(
                    f"{previous_extracted_sep_text}{extracted_sep_text}", is_last=True
                )
            )
        new_index = len(line_to_parse)
        return trailing_whitespace, new_index

    @staticmethod
    def __parse_table_3b(
        tr_columns: List[TableColumn],
        extracted_sep_text: str,
        after_sep_index: int,
        previous_extracted_sep_text: str,
        line_to_parse: str,
    ) -> Tuple[str, int]:
        previous_extracted_sep_text += f"{extracted_sep_text}|"
        new_index = after_sep_index + 1
        if new_index >= len(line_to_parse):
            extracted_sep_text = f"{previous_extracted_sep_text}|"
            previous_extracted_sep_text = ""
            first_non_matching_index, _ = ParserHelper.collect_while_spaces(
                extracted_sep_text, 0
            )
            assert first_non_matching_index != len(extracted_sep_text)
            tr_columns.append(TableColumn(extracted_sep_text, is_last=True))
        return previous_extracted_sep_text, new_index

    @staticmethod
    def __parse_table_3c(
        previous_extracted_sep_text: str,
        extracted_sep_text: str,
        tr_columns: List[TableColumn],
        trailing_whitespace: str,
    ) -> Tuple[str, str]:
        extracted_sep_text = f"{previous_extracted_sep_text}{extracted_sep_text}"
        if extracted_sep_text:
            first_non_matching_index, _ = ParserHelper.collect_while_spaces(
                extracted_sep_text, 0
            )
            if first_non_matching_index == len(extracted_sep_text):
                trailing_whitespace = extracted_sep_text
            else:
                tr_columns.append(TableColumn(extracted_sep_text, is_last=True))
        return extracted_sep_text, trailing_whitespace

    @staticmethod
    def __parse_table_3(
        line_to_parse: str, column_separator_characters: str, new_index: int
    ) -> Tuple[List[TableColumn], str, str, int]:
        tr_columns: List[TableColumn] = []
        trailing_whitespace = ""
        previous_extracted_sep_text = ""
        while new_index < len(line_to_parse):
            after_sep_index, extracted_sep_text = (
                ParserHelper.collect_until_one_of_characters(
                    line_to_parse, new_index, column_separator_characters
                )
            )
            assert after_sep_index is not None
            assert extracted_sep_text is not None
            if after_sep_index >= len(line_to_parse):
                trailing_whitespace, new_index = TableParseHelper.__parse_table_3a(
                    tr_columns,
                    extracted_sep_text,
                    trailing_whitespace,
                    previous_extracted_sep_text,
                    line_to_parse,
                )
                continue
            if (
                after_sep_index > 0
                and line_to_parse[after_sep_index] == "|"
                and line_to_parse[after_sep_index - 1] == "\\"
            ):
                previous_extracted_sep_text, new_index = (
                    TableParseHelper.__parse_table_3b(
                        tr_columns,
                        extracted_sep_text,
                        after_sep_index,
                        previous_extracted_sep_text,
                        line_to_parse,
                    )
                )
                continue
            new_index = after_sep_index + 1
            if line_to_parse[after_sep_index] == "\n":
                extracted_sep_text, trailing_whitespace = (
                    TableParseHelper.__parse_table_3c(
                        previous_extracted_sep_text,
                        extracted_sep_text,
                        tr_columns,
                        trailing_whitespace,
                    )
                )
                break
            tr_columns.append(
                TableColumn(f"{previous_extracted_sep_text}{extracted_sep_text}|")
            )
            previous_extracted_sep_text = ""
        return tr_columns, trailing_whitespace, previous_extracted_sep_text, new_index

    @staticmethod
    def __parse_table_2(
        table_rows: List[TableRow], col_as: List[Optional[str]]
    ) -> bool:
        if len(table_rows[0].columns) != len(table_rows[1].columns):
            return False
        for next_column in table_rows[1].columns:
            column_text = next_column.text
            text_index = 0
            has_lead = has_trail = False
            if ParserHelper.is_character_at_index(column_text, text_index, ":"):
                has_lead = True
                text_index += 1
            dashes_collected, dashes_new_index = ParserHelper.collect_while_character(
                column_text, text_index, "-"
            )
            assert dashes_collected is not None
            assert dashes_new_index is not None
            if dashes_collected and dashes_new_index <= len(column_text):
                if ParserHelper.is_character_at_index(
                    column_text, dashes_new_index, ":"
                ):
                    has_trail = True
                    dashes_new_index += 1
                is_complete = len(column_text) == dashes_new_index
                if is_complete:
                    column_alignment = None
                    if has_lead:
                        column_alignment = "center" if has_trail else "left"
                    elif has_trail:
                        column_alignment = "right"
                    col_as.append(column_alignment)
            else:
                is_complete = False
            if not is_complete:
                return False
        return True

    @staticmethod
    def __create_table_token(
        new_index: int,
        xyz: List[TableRow],
        col_as: List[Optional[str]],
    ) -> Tuple[bool, int, Optional[TableTuple]]:
        # POGGER.debug(
        #     ">>collected_destination(normalized)>>$",
        #     normalized_destination,
        # )

        # if (
        #     not inline_title
        #     and line_title_whitespace
        #     and line_title_whitespace[-1] == ParserHelper.newline_character
        # ):
        #     line_title_whitespace = line_title_whitespace[:-1]

        # POGGER.debug(">>inline_link>>$<<", inline_link)
        # POGGER.debug(">>inline_title>>$<<", inline_title)
        parsed_lrd_tuple = TableTuple(
            normalized_destination="",
            xyz=xyz,
            col_as=col_as,
            # LinkReferenceTitles(inline_link, inline_title),
            # LinkReferenceInfo(
            #     collected_destination,
            #     line_destination_whitespace,
            #     inline_raw_link,
            #     line_title_whitespace,
            #     inline_raw_title,
            #     end_whitespace,
            # ),
        )
        return True, new_index, parsed_lrd_tuple

    @staticmethod
    def __is_table_start(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
    ) -> bool:
        """
        Determine whether or not we have a valid table start.  We can tell this because
        a proper table will have a pipe character `|` at least once in every line.
        """

        if parser_state.token_stack[-1].is_paragraph:
            return False

        POGGER.debug(
            "__is_table_start - extracted_whitespace:>:$:<",
            extracted_whitespace,
        )
        POGGER.debug("__is_table_start - line_to_parse:>:$:<", line_to_parse)
        if (
            TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
        ) and TableParseHelper.__table_column_separator_character in line_to_parse:
            POGGER.debug("__is_table_start - potential")

            # remaining_line, continue_with_lrd = line_to_parse[start_index + 1 :], True
            # if (
            #     remaining_line
            #     and remaining_line[-1] == InlineBackslashHelper.backslash_character
            # ):
            #     remaining_line_size, start_index, found_index = (
            #         len(remaining_line),
            #         0,
            #         remaining_line.find(
            #             InlineBackslashHelper.backslash_character, start_index
            #         ),
            #     )
            #     POGGER.debug(">>$<<$", remaining_line, remaining_line_size)
            #     POGGER.debug(">>$<<$", remaining_line, start_index)
            #     POGGER.debug(">>$<<", found_index)
            #     while found_index != -1 and found_index < (remaining_line_size - 1):
            #         start_index = found_index + 2
            #         POGGER.debug(">>$<<$", remaining_line, start_index)
            #         found_index = remaining_line.find(
            #             InlineBackslashHelper.backslash_character, start_index
            #         )
            #         POGGER.debug(">>$<<", found_index)
            #     POGGER.debug(">>>>>>>$<<", found_index)
            #     continue_with_lrd = found_index != remaining_line_size - 1
            return True  # continue_with_lrd
        return False


## Different from LRD.



"""
Module to help contain the elements of a link.
"""

# pylint: disable=too-many-instance-attributes
from dataclasses import dataclass
from typing import Optional


@dataclass
class LinkHelperProperties:
    """
    Class to help contain the elements of a link.
    """

    ex_label: Optional[str]
    label_type: Optional[str]
    before_link_whitespace: Optional[str]
    inline_link: Optional[str]
    pre_inline_link: Optional[str]
    before_title_whitespace: Optional[str]
    bounding_character: Optional[str]
    inline_title: Optional[str]
    pre_inline_title: Optional[str]
    after_title_whitespace: Optional[str]
    did_use_angle_start: Optional[bool]

    def __init__(self) -> None:
        self.ex_label = None
        self.label_type = None
        self.did_use_angle_start = False
        self.before_link_whitespace = None
        self.inline_link = None
        self.pre_inline_link = None
        self.before_title_whitespace = None
        self.bounding_character = None
        self.inline_title = None
        self.pre_inline_title = None
        self.after_title_whitespace = None


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks for trailing punctuation in headings.
"""

from typing import List, Optional, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd026(RulePlugin):
    """
    Class to implement a plugin that looks for trailing punctuation in headings.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__start_token: Optional[MarkdownToken] = None
        self.__heading_text = ""
        self.__punctuation = ""

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="no-trailing-punctuation",
            plugin_id="MD026",
            plugin_enabled_by_default=True,
            plugin_description="Trailing punctuation present in heading text.",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md026.md",
            plugin_configuration="punctuation",
        )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__punctuation = self.plugin_configuration.get_string_property_with_default(
            "punctuation", ".,;:!。，；：！"
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [QueryConfigItem("punctuation", self.__punctuation)]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__start_token = None
        self.__heading_text = ""

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_setext_heading or token.is_atx_heading:
            self.__heading_text = ""
            self.__start_token = token
        elif token.is_end_token:
            self.__next_token_end(token, context)
        elif self.__start_token:
            if token.is_text:
                text_token = cast(TextMarkdownToken, token)
                self.__heading_text += text_token.token_text
            else:
                self.__heading_text = ""

    def __next_token_end(
        self, token: MarkdownToken, context: PluginScanContext
    ) -> None:
        if token.is_setext_heading_end or token.is_atx_heading_end:
            if self.__heading_text:
                if token.is_atx_heading_end:
                    use_original_position = False
                    line_delta = 0
                    column_delta = len(self.__heading_text) - 1
                else:
                    use_original_position = True
                    line_delta = self.__heading_text.count(
                        ParserHelper.newline_character
                    )
                    if line_delta:
                        split_heading_text = self.__heading_text.split(
                            ParserHelper.newline_character
                        )
                        column_delta = len(split_heading_text[-1]) - 1
                    else:
                        column_delta = len(self.__heading_text) - 1
                if self.__heading_text[-1] in self.__punctuation:
                    assert self.__start_token is not None
                    self.report_next_token_error(
                        context,
                        self.__start_token,
                        line_number_delta=line_delta,
                        column_number_delta=column_delta,
                        use_original_position=use_original_position,
                    )
            self.__start_token = None
        else:
            self.__heading_text = ""



"""
Module to provide processing for the leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.blank_line_markdown_token import BlankLineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.stack_token import (
    BlockQuoteStackToken,
    ListStackToken,
    ParagraphStackToken,
)
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class LeafBlockProcessorParagraph:
    """
    Class to provide processing for the leaf blocks.
    """

    # pylint: disable=too-many-arguments
    @staticmethod
    def parse_paragraph(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        block_quote_data: BlockQuoteData,
        text_removed_by_container: Optional[str],
        original_line: str,
    ) -> List[MarkdownToken]:
        """
        Handle the parsing of a paragraph.
        """
        POGGER.debug(">>text_removed_by_container>>:$:<<", text_removed_by_container)
        POGGER.debug(">>extracted_whitespace>>:$:<<", extracted_whitespace)
        POGGER.debug("position_marker.text_to_parse=:$:", position_marker.text_to_parse)
        POGGER.debug("position_marker.index_number=:$:", position_marker.index_number)
        POGGER.debug("position_marker.index_indent=:$:", position_marker.index_indent)

        if parser_state.no_para_start_if_empty and position_marker.index_number >= len(
            position_marker.text_to_parse
        ):
            POGGER.debug("Escaping paragraph due to empty w/ blank")
            return [
                BlankLineMarkdownToken(
                    extracted_whitespace, position_marker, len(extracted_whitespace)
                )
            ]

        POGGER.debug(
            "parse_paragraph>block_quote_data.stack_count>$>block_quote_data.current_count>$<",
            block_quote_data.stack_count,
            block_quote_data.current_count,
        )

        container_index, adjusted_whitespace_length = (
            parser_state.find_last_container_on_stack(),
            0,
        )
        if container_index > 0:
            adjusted_whitespace_length = (
                LeafBlockProcessorParagraph.__adjust_paragraph_for_containers(
                    parser_state,
                    container_index,
                    extracted_whitespace,
                    adjusted_whitespace_length,
                )
            )

        POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)
        POGGER.debug("adjusted_whitespace_length=:$:", adjusted_whitespace_length)
        (
            new_tokens,
            extracted_whitespace,
            did_add_paragraph_token,
        ) = LeafBlockProcessorParagraph.__handle_paragraph_prep(
            parser_state,
            block_quote_data,
            position_marker,
            extracted_whitespace,
            original_line,
        )
        POGGER.debug("new_tokens=:$:", new_tokens)
        POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)
        POGGER.debug("did_add_paragraph_token=:$:", did_add_paragraph_token)
        POGGER.debug(
            "position_marker.text_to_parse=:$:",
            position_marker.text_to_parse[position_marker.index_number :],
        )

        adjusted_index = position_marker.index_number
        if did_add_paragraph_token and ParserHelper.is_character_at_index_one_of(
            position_marker.text_to_parse, adjusted_index, Constants.ascii_whitespace
        ):
            POGGER.debug(
                "position_marker.text_to_parse=:$:", position_marker.text_to_parse
            )
            POGGER.debug("adjusted_index=:$:", adjusted_index)
            adjusted_index, extracted_whitespace = (
                ParserHelper.extract_ascii_whitespace_verified(
                    position_marker.text_to_parse, adjusted_index
                )
            )
            POGGER.debug("adjusted_index=:$:", adjusted_index)
            POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)

        new_tokens.append(
            LeafBlockProcessorParagraph.__parse_paragraph_create_text_token(
                parser_state,
                position_marker,
                adjusted_index,
                extracted_whitespace,
                original_line,
            )
        )
        return new_tokens

    # pylint: enable=too-many-arguments

    @staticmethod
    def __parse_paragraph_create_text_token(
        parser_state: ParserState,
        position_marker: PositionMarker,
        adjusted_index: int,
        extracted_whitespace: str,
        original_line: str,
    ) -> TextMarkdownToken:
        text_to_parse = position_marker.text_to_parse[adjusted_index:]
        POGGER.debug("--add-text-token--")
        POGGER.debug("text_to_parse=:$:", text_to_parse)
        POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)
        POGGER.debug("original_line=:$:", original_line)

        (
            corrected_tab_text,
            corrected_extracted_whitespace,
        ) = LeafBlockProcessorParagraph.__calculate_corrected_tab_text(
            parser_state,
            original_line,
            text_to_parse,
            extracted_whitespace,
            adjusted_index,
        )
        POGGER.debug("corrected_tab_text=:$:", corrected_tab_text)
        POGGER.debug(
            "corrected_extracted_whitespace=:$:", corrected_extracted_whitespace
        )

        return TextMarkdownToken(
            text_to_parse,
            corrected_extracted_whitespace,
            position_marker=position_marker,
            tabified_text=corrected_tab_text,
        )

    @staticmethod
    def __calculate_corrected_tab_text(
        parser_state: ParserState,
        original_line: str,
        text_to_parse: str,
        extracted_whitespace: str,
        adjusted_index: int,
    ) -> Tuple[str, str]:
        corrected_tab_text = ""
        if ParserHelper.tab_character in original_line:
            (
                corrected_extracted_whitespace,
                original_line,
                checked_whitespace_for_tab,
                is_block_quote_container,
            ) = LeafBlockProcessorParagraph.__calculate_corrected_tab_text_prefix(
                parser_state, extracted_whitespace, text_to_parse, original_line
            )
            POGGER.debug("checked_whitespace_for_tab=:$:", checked_whitespace_for_tab)
            POGGER.debug("is_block_quote_container=:$:", is_block_quote_container)

            if original_line.endswith(text_to_parse):
                corrected_index = len(original_line) - len(text_to_parse)
                assert corrected_index >= 0, "Index cannot be negative."
                corrected_tab_text = original_line[corrected_index:]
                assert (
                    corrected_tab_text == text_to_parse
                ), "Constructed string must equal provided text."
            else:
                POGGER.debug("original_line=:$:", original_line)
                POGGER.debug("text_to_parse=:$:", text_to_parse)
                initial_offset = (
                    adjusted_index
                    if checked_whitespace_for_tab and not is_block_quote_container
                    else 0
                )
                if checked_whitespace_for_tab and not is_block_quote_container:
                    detabified_whitespace = TabHelper.detabify_string(
                        corrected_extracted_whitespace
                    )
                    initial_offset = len(detabified_whitespace)
                POGGER.debug("initial_offset=:$:", initial_offset)
                corrected_tab_text, _, _ = TabHelper.find_detabify_string_verified(
                    original_line,
                    text_to_parse,
                    use_proper_traverse=True,
                    initial_offset=initial_offset,
                )
                POGGER.debug("adj_text_to_parse=:$:", corrected_tab_text)
                corrected_index = 0
        else:
            corrected_extracted_whitespace = extracted_whitespace
        return corrected_tab_text, corrected_extracted_whitespace

    # pylint: disable=too-many-locals
    @staticmethod
    def __calculate_corrected_tab_text_prefix(
        parser_state: ParserState,
        extracted_whitespace: str,
        text_to_parse: str,
        original_line: str,
    ) -> Tuple[str, str, bool, bool]:
        POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)
        split_tab_with_block_quote_suffix = True
        checked_whitespace_for_tab = False
        is_block_quote_container = False
        if extracted_whitespace:
            first_non_whitespace_character = text_to_parse[0]
            first_non_whitespace_character_index = original_line.find(
                first_non_whitespace_character
            )
            corrected_extracted_whitespace = original_line[
                :first_non_whitespace_character_index
            ]
            POGGER.debug(
                "corrected_extracted_whitespace=:$:", corrected_extracted_whitespace
            )
            last_container_index = parser_state.find_last_container_on_stack()
            last_container_token = parser_state.token_stack[last_container_index]
            is_block_quote_container = last_container_token.is_block_quote
            (
                corrected_prefix,
                corrected_suffix,
                split_tab,
                split_tab_with_block_quote_suffix,
            ) = TabHelper.match_tabbed_whitespace(
                extracted_whitespace, corrected_extracted_whitespace
            )
            checked_whitespace_for_tab = True
            POGGER.debug("split_tab=:$:", split_tab)
            POGGER.debug(
                "split_tab_with_block_quote_suffix=:$:",
                split_tab_with_block_quote_suffix,
            )
            POGGER.debug("extracted_whitespace=:$:", extracted_whitespace)
            if split_tab:
                if split_tab_with_block_quote_suffix:
                    TabHelper.adjust_block_quote_indent_for_tab(parser_state)
                else:
                    alternate_leading_space = (
                        corrected_prefix
                        if corrected_prefix and "\t" in corrected_prefix
                        else None
                    )
                    TabHelper.adjust_block_quote_indent_for_tab(
                        parser_state,
                        corrected_prefix + corrected_suffix,
                        original_line=original_line,
                        alternate_list_leading_space=alternate_leading_space,
                    )

            corrected_extracted_whitespace = corrected_suffix
            original_line = original_line[first_non_whitespace_character_index:]
        else:
            corrected_extracted_whitespace = extracted_whitespace
        return (
            corrected_extracted_whitespace,
            original_line,
            checked_whitespace_for_tab,
            is_block_quote_container,
        )

    # pylint: enable=too-many-locals

    @staticmethod
    def __adjust_paragraph_for_containers(
        parser_state: ParserState,
        container_index: int,
        extracted_whitespace: str,
        adjusted_whitespace_length: int,
    ) -> int:
        if not parser_state.token_stack[container_index].is_block_quote:
            top_list_token = cast(
                ListStackToken, parser_state.token_stack[container_index]
            )
            POGGER.debug(">>list-owners>>$", top_list_token)
            adjusted_whitespace_length = (
                LeafBlockProcessorParagraph.__adjust_paragraph_for_list(
                    top_list_token, extracted_whitespace
                )
            )
        POGGER.debug(">>adjusted_whitespace_length>>$", adjusted_whitespace_length)
        return adjusted_whitespace_length

    @staticmethod
    def __adjust_paragraph_for_list(
        top_list_token: ListStackToken, extracted_whitespace: str
    ) -> int:
        ex_ws_length = len(extracted_whitespace)
        POGGER.debug(">>owners-indent>>$", top_list_token.indent_level)
        POGGER.debug(">>ws_before_marker>>$", top_list_token.ws_before_marker)
        POGGER.debug(">>ws_after_marker>>$", top_list_token.ws_after_marker)
        POGGER.debug(">>last_new_list_token>>$", top_list_token.last_new_list_token)
        POGGER.debug(">>extracted_whitespace>>$", ex_ws_length)

        dominant_indent, original_list_indent, indent_delta = (
            (
                top_list_token.last_new_list_token.indent_level
                if top_list_token.last_new_list_token
                else top_list_token.indent_level
            ),
            top_list_token.indent_level - 2,
            (
                top_list_token.ws_after_marker - 1
                if top_list_token.ws_after_marker > 1
                else 0
            ),
        )
        POGGER.debug(">>dominant_indent>>$>>", dominant_indent)
        original_text_indent = (
            ex_ws_length
            + top_list_token.indent_level
            - top_list_token.ws_before_marker
            - indent_delta
        )
        POGGER.debug(">>original_list_indent>>$>>", original_list_indent)
        POGGER.debug(">>original_text_indent>$>>", original_text_indent)
        return (
            dominant_indent - original_text_indent
            if dominant_indent > original_text_indent >= 4
            else 0
        )

    @staticmethod
    def __handle_paragraph_prep_block_reduce(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
    ) -> List[MarkdownToken]:
        new_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            only_these_blocks=[BlockQuoteStackToken],
            include_block_quotes=True,
        )
        _ = (position_marker, extracted_whitespace)
        # if new_tokens:
        # assert parser_state.token_stack[-1].is_list, "Must be within list block."
        # search_index = len(parser_state.token_stack)
        # leading_space_length = (
        #     len(extracted_whitespace) + position_marker.index_indent
        # )
        # did_once = False
        # while parser_state.token_stack[search_index - 1].is_list:
        #     list_token = cast(
        #         ListStackToken, parser_state.token_stack[search_index - 1]
        #     )
        #     if list_token.indent_level <= leading_space_length:
        #         break
        #     search_index -= 1
        #     did_once = True
        # if did_once:
        #     # POGGER.debug("lsl $", parser_state.token_stack[search_index])
        #     end_token = cast(EndMarkdownToken, new_tokens[-1])
        #     end_token.set_extra_end_data(None)

        #     (
        #         container_level_tokens,
        #         _,
        #     ) = parser_state.close_open_blocks_fn(
        #         parser_state,
        #         until_this_index=search_index,
        #         include_lists=True,
        #         caller_can_handle_requeue=False,
        #         requeue_reset=True,
        #     )
        #     new_tokens.extend(container_level_tokens)
        return new_tokens

    @staticmethod
    def __handle_paragraph_prep(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        original_line: str,
    ) -> Tuple[List[MarkdownToken], str, bool]:
        # In cases where the list ended on the same line as we are processing, the
        # container tokens will not yet be added to the token_document.  As such,
        # make sure to construct a "proper" list that takes those into account
        # before checking to see if this is an issue.
        adjusted_document = parser_state.token_document[:]
        assert (
            parser_state.same_line_container_tokens is not None
        ), "Should have some tokens."
        adjusted_document.extend(parser_state.same_line_container_tokens)
        did_add_paragraph_token = False

        if (
            len(adjusted_document) >= 2
            and adjusted_document[-1].is_blank_line
            and adjusted_document[-2].is_any_list_token
        ):
            (
                _,
                last_list_index,
            ) = LeafBlockProcessorParagraph.verify_list_in_process(parser_state)
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state, until_this_index=last_list_index
            )
        elif block_quote_data.stack_count != 0 and block_quote_data.current_count == 0:
            new_tokens = (
                LeafBlockProcessorParagraph.__handle_paragraph_prep_block_reduce(
                    parser_state, position_marker, extracted_whitespace
                )
            )
        else:
            new_tokens = []

        if not parser_state.token_stack[-1].is_paragraph:
            if ParserHelper.tab_character in original_line:
                extracted_whitespace = (
                    LeafBlockProcessorParagraph.__paragraph_prep_whitespace_with_tab(
                        parser_state,
                        position_marker,
                        original_line,
                        extracted_whitespace,
                    )
                )

            new_paragraph_token = ParagraphMarkdownToken(
                extracted_whitespace, position_marker
            )
            parser_state.token_stack.append(ParagraphStackToken(new_paragraph_token))
            new_tokens.append(new_paragraph_token)
            extracted_whitespace = ""
            did_add_paragraph_token = True
        return new_tokens, extracted_whitespace, did_add_paragraph_token

    @staticmethod
    def __paragraph_prep_whitespace_with_tab(
        parser_state: ParserState,
        position_marker: PositionMarker,
        original_line: str,
        extracted_whitespace: str,
    ) -> str:
        POGGER.debug(">>text_to_parse>>$>>", position_marker.text_to_parse)
        POGGER.debug(">>index_number>>$>>", position_marker.index_number)

        POGGER.debug(">>original_line>:$:<", original_line)
        POGGER.debug(
            ">>position_marker.text_to_parse[position_marker.index_number:]>:$:<",
            position_marker.text_to_parse[position_marker.index_number :],
        )
        (
            _,
            _,
            rest_of_string_index,
        ) = TabHelper.find_detabify_string(
            original_line,
            position_marker.text_to_parse[position_marker.index_number :],
            use_proper_traverse=True,
        )
        POGGER.debug(">>rest_of_string_index>>$>>", rest_of_string_index)
        prefix = original_line[:rest_of_string_index]
        if prefix and ParserHelper.tab_character in prefix:
            POGGER.debug(">>extracted_whitespace>:$:<", extracted_whitespace)
            POGGER.debug(">>prefix>:$:<", prefix)
            (
                corrected_prefix,
                extracted_whitespace,
                split_tab,
                split_tab_with_block_quote_suffix,
            ) = TabHelper.match_tabbed_whitespace(extracted_whitespace, prefix)
            POGGER.debug(">>corrected_prefix>:$:<", corrected_prefix)
            POGGER.debug(">>corrected_suffix>:$:<", extracted_whitespace)
            POGGER.debug(">>split_tab>>$>>", split_tab)
            if split_tab:
                if split_tab_with_block_quote_suffix:
                    TabHelper.adjust_block_quote_indent_for_tab(parser_state)
                else:
                    alternate_leading_space = (
                        corrected_prefix
                        if corrected_prefix and "\t" in corrected_prefix
                        else None
                    )
                    TabHelper.adjust_block_quote_indent_for_tab(
                        parser_state,
                        corrected_prefix + extracted_whitespace,
                        original_line=original_line,
                        alternate_list_leading_space=alternate_leading_space,
                    )
        return extracted_whitespace

    @staticmethod
    def check_for_list_in_process(parser_state: ParserState) -> Tuple[bool, int]:
        """
        From the end of the stack, check to see if there is already a list in progress.
        """

        stack_index = len(parser_state.token_stack) - 1
        while stack_index >= 0 and not parser_state.token_stack[stack_index].is_list:
            stack_index -= 1
        return stack_index >= 0, stack_index

    @staticmethod
    def verify_list_in_process(parser_state: ParserState) -> Tuple[bool, int]:
        """
        From the end of the stack, verify that there is already a list in progress.
        """
        did_find_list, stack_index = (
            LeafBlockProcessorParagraph.check_for_list_in_process(parser_state)
        )
        assert did_find_list, "List must exist on the stack."
        return did_find_list, stack_index



"""
Module to implement a plugin that looks for bare URLs in the files.
"""

from typing import cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class RuleMd034(RulePlugin):
    """
    Class to implement a plugin that looks for bare URLs in the files.
    """

    __valid_uri_types = ["http:", "https:", "ftp:", "ftps:"]

    def __init__(self) -> None:
        super().__init__()
        self.__in_code_block: bool = False
        self.__in_html_block: bool = False
        self.__in_link: bool = False

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-bare-urls",
            plugin_id="MD034",
            plugin_enabled_by_default=True,
            plugin_description="Bare URL used",
            plugin_version="0.5.1",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md034.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__in_code_block = False
        self.__in_html_block = False
        self.__in_link = False

    # pylint: disable=too-many-arguments
    def __evaluate_possible_url(
        self,
        source_text: str,
        url_prefix: str,
        found_index: int,
        context: PluginScanContext,
        token: MarkdownToken,
    ) -> None:
        if found_index == 0 or source_text[found_index - 1] in (
            " ",
            ParserHelper.newline_character,
        ):
            url_start_sequence = source_text[found_index + len(url_prefix) :]
            if (
                len(url_start_sequence) >= 3
                and url_start_sequence.startswith("//")
                and url_start_sequence[2] not in (" ", ParserHelper.newline_character)
            ):
                (
                    column_number_delta,
                    line_number_delta,
                ) = ParserHelper.adjust_for_newlines(source_text, 0, found_index)
                self.report_next_token_error(
                    context,
                    token,
                    line_number_delta=line_number_delta
                    + context.calc_pragma_offset(token, line_number_delta),
                    column_number_delta=column_number_delta,
                )

    # pylint: enable=too-many-arguments

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if (
            token.is_text
            and not self.__in_code_block
            and not self.__in_html_block
            and not self.__in_link
        ):
            text_token = cast(TextMarkdownToken, token)
            for url_prefix in RuleMd034.__valid_uri_types:
                start_index = 0
                found_index = text_token.token_text.find(url_prefix, start_index)
                while found_index != -1:
                    self.__evaluate_possible_url(
                        text_token.token_text, url_prefix, found_index, context, token
                    )
                    start_index = found_index + len(url_prefix)
                    found_index = text_token.token_text.find(url_prefix, start_index)

        elif token.is_code_block:
            self.__in_code_block = True
        elif token.is_code_block_end:
            self.__in_code_block = False
        elif token.is_html_block:
            self.__in_html_block = True
        elif token.is_html_block_end:
            self.__in_html_block = False
        elif token.is_inline_link:
            self.__in_link = True
        elif token.is_inline_link_end:
            self.__in_link = False



"""
Module to help with the parsing of text inline elements.
"""

import logging
from typing import List, Optional, Set, Tuple, cast

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.inline.emphasis_helper import EmphasisHelper
from pymarkdown.inline.inline_handler_helper import InlineHandlerHelper
from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.inline.inline_line_end_helper import InlineLineEndHelper
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse
from pymarkdown.inline.inline_tabified_text_block_helper import (
    InlineTabifiedTextBlockHelper,
)
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


# pylint: disable=too-few-public-methods, too-many-lines


class InlineTextBlockHelper:
    """
    Class to help with the parsing of text inline elements.
    """

    @staticmethod
    def __process_inline_text_block_prepare(
        source_text: str,
        whitespace_to_recombine: Optional[str],
        is_para: bool,
        is_setext: bool,
        para_space: Optional[str],
    ) -> Tuple[str, Optional[List[str]]]:

        if whitespace_to_recombine:
            source_text, _ = ParserHelper.recombine_string_with_whitespace(
                source_text, whitespace_to_recombine
            )

        split_para_space: Optional[List[str]] = None
        if is_para or is_setext:
            assert (
                para_space is not None
            ), "If in a paragraph or setext text, para_space must be defined."
            split_para_space = para_space.split(ParserHelper.newline_character)

        return source_text, split_para_space

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_inline_text_block_start(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        whitespace_to_recombine: Optional[str],
        is_para: bool,
        is_setext: bool,
        para_space: Optional[str],
        start_index: int,
    ) -> Tuple[int, Set[int], str, Optional[List[str]]]:
        pragma_line_numbers: Set[int] = set(list(parser_properties.pragma_lines.keys()))

        source_text, split_para_space = (
            InlineTextBlockHelper.__process_inline_text_block_prepare(
                source_text, whitespace_to_recombine, is_para, is_setext, para_space
            )
        )

        next_index = ParserHelper.index_any_of(
            source_text,
            InlineHandlerHelper.valid_inline_text_block_sequence_starts,
            start_index,
        )
        return next_index, pragma_line_numbers, source_text, split_para_space

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod  # noqa: C901
    def process_inline_text_block(  # noqa: C901
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        coalesced_stack: List[MarkdownToken],
        starting_whitespace: str = "",
        whitespace_to_recombine: Optional[str] = None,
        is_setext: bool = False,
        is_para: bool = False,
        para_space: Optional[str] = None,
        line_number: int = 0,
        column_number: int = 0,
        para_owner: Optional[ParagraphMarkdownToken] = None,
        tabified_text: Optional[str] = None,
        is_in_table: bool = False,
    ) -> List[MarkdownToken]:
        """
        Process a text block for any inline items.

        Note: This is one of the more heavily traffic functions in the
        parser.  Debugging should be uncommented only if needed.
        """

        (
            last_line_number,
            last_column_number,
            current_string,
            current_string_unresolved,
        ) = (line_number, column_number, "", "")
        start_index = 0
        newlines_encountered = 0
        inline_blocks: List[MarkdownToken] = []
        end_string: Optional[str] = ""

        next_index, pragma_line_numbers, source_text, split_para_space = (
            InlineTextBlockHelper.__process_inline_text_block_start(
                parser_properties,
                source_text,
                whitespace_to_recombine,
                is_para,
                is_setext,
                para_space,
                start_index,
            )
        )
        while next_index != -1:
            old_next_index = next_index
            (
                line_number,
                column_number,
                end_string,
                current_string,
                current_string_unresolved,
                starting_whitespace,
                split_para_space,
                last_line_number,
                last_column_number,
                start_index,
                next_index,
                adj_newlines,
            ) = InlineTextBlockHelper.__handle_next_inline_character(
                parser_properties,
                source_text,
                start_index,
                next_index,
                inline_blocks,
                current_string,
                current_string_unresolved,
                line_number,
                column_number,
                para_owner,
                coalesced_stack,
                end_string,
                is_setext,
                whitespace_to_recombine,
                starting_whitespace,
                last_line_number,
                last_column_number,
                split_para_space,
                tabified_text,
                newlines_encountered,
                is_in_table,
                pragma_line_numbers,
                para_space,
            )
            if adj_newlines:
                newlines_encountered += adj_newlines
            if source_text[old_next_index] == "\n":
                newlines_encountered += 1

        return InlineTextBlockHelper.__complete_inline_block_processing(
            inline_blocks,
            source_text,
            start_index,
            current_string,
            end_string,
            starting_whitespace,
            is_setext,
            last_line_number,
            last_column_number,
            tabified_text,
            newlines_encountered,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __handle_next_inline_character(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        start_index: int,
        next_index: int,
        inline_blocks: List[MarkdownToken],
        current_string: str,
        current_string_unresolved: str,
        line_number: int,
        column_number: int,
        para_owner: Optional[ParagraphMarkdownToken],
        coalesced_stack: List[MarkdownToken],
        end_string: Optional[str],
        is_setext: bool,
        whitespace_to_recombine: Optional[str],
        starting_whitespace: str,
        last_line_number: int,
        last_column_number: int,
        split_para_space: Optional[List[str]],
        tabified_text: Optional[str],
        newlines_encountered: int,
        is_in_table: bool,
        pragma_line_numbers: Set[int],
        para_space: Optional[str] = None,
    ) -> Tuple[
        int,
        int,
        Optional[str],
        str,
        str,
        str,
        Optional[List[str]],
        int,
        int,
        int,
        int,
        int,
    ]:
        (
            reset_current_string,
            remaining_line,
            old_inline_blocks_count,
            old_inline_blocks_last_token,
            tabified_remaining_line,
            inline_request,
        ) = InlineTextBlockHelper.__handle_next_inline_character_setup(
            source_text,
            start_index,
            next_index,
            inline_blocks,
            tabified_text,
            newlines_encountered,
            current_string_unresolved,
            line_number,
            column_number,
            para_owner,
            parser_properties,
            coalesced_stack,
            current_string,
            whitespace_to_recombine,
            is_in_table,
            para_space,
        )

        (
            inline_response,
            line_number,
            column_number,
            was_column_number_reset,
            did_line_number_change,
            whitespace_to_add,
            remaining_line,
            end_string,
            current_string,
            current_string_unresolved,
            was_new_line,
            tabified_remaining_line,
        ) = InlineTextBlockHelper.__handle_next_special_character(
            parser_properties,
            source_text,
            next_index,
            inline_request,
            line_number,
            column_number,
            coalesced_stack,
            InlineResponse(),
            remaining_line,
            end_string,
            current_string,
            current_string_unresolved,
            inline_blocks,
            is_setext,
            whitespace_to_recombine,
            para_owner,
            tabified_text,
            tabified_remaining_line,
            pragma_line_numbers,
        )

        (
            reset_current_string,
            remaining_line,
            end_string,
            current_string,
            current_string_unresolved,
        ) = InlineTextBlockHelper.__cleanup_after_handling(
            inline_response,
            current_string,
            current_string_unresolved,
            remaining_line,
            tabified_remaining_line,
            reset_current_string,
            end_string,
        )

        (
            reset_current_string,
            starting_whitespace,
            end_string,
        ) = InlineTextBlockHelper.__create_new_text_token(
            inline_response,
            current_string,
            inline_blocks,
            starting_whitespace,
            end_string,
            last_line_number,
            last_column_number,
            reset_current_string,
        )

        (
            line_number,
            column_number,
            split_para_space,
            current_string,
            current_string_unresolved,
            last_line_number,
            last_column_number,
            new_start_index,
            next_index,
            end_string,
        ) = InlineTextBlockHelper.__handle_next_inline_character_finish_handling(
            line_number,
            column_number,
            split_para_space,
            was_new_line,
            coalesced_stack,
            remaining_line,
            did_line_number_change,
            was_column_number_reset,
            reset_current_string,
            inline_blocks,
            old_inline_blocks_count,
            old_inline_blocks_last_token,
            source_text,
            whitespace_to_add,
            inline_response,
            current_string,
            current_string_unresolved,
            last_line_number,
            last_column_number,
            end_string,
            pragma_line_numbers,
        )

        return (
            line_number,
            column_number,
            end_string,
            current_string,
            current_string_unresolved,
            starting_whitespace,
            split_para_space,
            last_line_number,
            last_column_number,
            new_start_index,
            next_index,
            inline_response.adj_newlines,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments,too-many-locals
    @staticmethod
    def __handle_next_inline_character_setup(
        source_text: str,
        start_index: int,
        next_index: int,
        inline_blocks: List[MarkdownToken],
        tabified_text: Optional[str],
        newlines_encountered: int,
        current_string_unresolved: str,
        line_number: int,
        column_number: int,
        para_owner: Optional[ParagraphMarkdownToken],
        parse_properties: ParseBlockPassProperties,
        coalesced_stack: List[MarkdownToken],
        current_string: str,
        whitespace_to_recombine: Optional[str],
        is_in_table: bool,
        para_space: Optional[str] = None,
    ) -> Tuple[bool, str, int, Optional[MarkdownToken], Optional[str], InlineRequest]:
        (
            remaining_line,
            old_inline_blocks_count,
            old_inline_blocks_last_token,
        ) = (
            source_text[start_index:next_index],
            len(inline_blocks),
            inline_blocks[-1] if inline_blocks else None,
        )
        tabified_remaining_line: Optional[str] = None
        if tabified_text:
            tabified_remaining_line = (
                InlineTabifiedTextBlockHelper.handle_next_inline_character_tabified(
                    source_text,
                    tabified_text,
                    newlines_encountered,
                    start_index,
                    next_index,
                )
            )

        last_container_token = coalesced_stack[-1] if coalesced_stack else None

        inline_request = InlineRequest(
            source_text,
            next_index,
            inline_blocks,
            remaining_line,
            tabified_remaining_line,
            current_string,
            current_string_unresolved,
            line_number,
            column_number,
            para_owner,
            tabified_text,
            parse_properties,
            last_container_token,
            whitespace_to_recombine,
            is_in_table,
            para_space,
        )
        return (
            False,
            remaining_line,
            old_inline_blocks_count,
            old_inline_blocks_last_token,
            tabified_remaining_line,
            inline_request,
        )

    # pylint: enable=too-many-arguments,too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __complete_inline_block_processing(
        inline_blocks: List[MarkdownToken],
        source_text: str,
        start_index: int,
        current_string: str,
        end_string: Optional[str],
        starting_whitespace: str,
        is_setext: bool,
        line_number: int,
        column_number: int,
        tabified_text: Optional[str],
        newlines_encountered: int,
    ) -> List[MarkdownToken]:
        POGGER.debug("__cibp>inline_blocks>$<", inline_blocks)
        POGGER.debug("__cibp>source_text>$<", source_text)
        POGGER.debug("__cibp>start_index>$<", start_index)
        POGGER.debug("__cibp>current_string>$<", current_string)
        POGGER.debug("__cibp>end_string>$<", end_string)
        POGGER.debug(
            "__cibp>starting_whitespace>$<",
            starting_whitespace,
        )
        POGGER.debug("__cibp>is_setext>$<", is_setext)
        POGGER.debug("__cibp>line_number>$<", line_number)
        POGGER.debug("__cibp>column_number>$<", column_number)

        if start_index < len(source_text):
            text_to_append = source_text[start_index:]
            POGGER.debug("text_to_append>:$:<", text_to_append)

            POGGER.debug("tabified_text=>:$:<", tabified_text)
            if tabified_text:
                text_to_append = InlineTabifiedTextBlockHelper.complete_inline_block_processing_tabified(
                    source_text,
                    start_index,
                    tabified_text,
                    newlines_encountered,
                    inline_blocks,
                    end_string,
                )
                POGGER.debug("text_to_append>:$:<", text_to_append)

            POGGER.debug("current_string>:$:<", current_string)
            POGGER.debug("text_to_append>:$:<", text_to_append)
            current_string = InlineHelper.append_text(current_string, text_to_append)
            POGGER.debug("current_string>:$:<", current_string)

        have_processed_once = inline_blocks or start_index != 0
        if current_string or not have_processed_once:
            InlineTextBlockHelper.__complete_inline_block_processing_build_token(
                current_string,
                end_string,
                starting_whitespace,
                is_setext,
                inline_blocks,
                line_number,
                column_number,
            )
        POGGER.debug(">>$<<", inline_blocks)

        EmphasisHelper.resolve_inline_emphasis(inline_blocks, None)
        return inline_blocks

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_next_special_character_remaining_reduce(
        inline_request: InlineRequest, left_to_consume: int
    ) -> None:
        inline_block_index = len(inline_request.inline_blocks) - 1
        stop_block_token = None
        while (
            stop_block_token is None
            and inline_block_index >= 0
            and (
                inline_request.inline_blocks[inline_block_index].is_text
                or inline_request.inline_blocks[inline_block_index].is_special_text
            )
            and left_to_consume > 0
        ):
            current_block_token = cast(
                TextMarkdownToken,
                inline_request.inline_blocks[inline_block_index],
            )
            token_text_len = len(current_block_token.token_text)
            if left_to_consume <= token_text_len:
                stop_block_token = current_block_token
            else:
                left_to_consume -= token_text_len
                inline_block_index -= 1
        assert (
            inline_block_index >= 0 and stop_block_token is not None
        ), "End of loop criteria."
        if len(stop_block_token.token_text) > left_to_consume:
            left_text = stop_block_token.token_text[:-left_to_consume]
            new_token = TextMarkdownToken(
                left_text,
                stop_block_token.extracted_whitespace,
                stop_block_token.end_whitespace,
                tabified_text=stop_block_token.tabified_text,
                line_number=stop_block_token.line_number,
                column_number=stop_block_token.column_number,
            )
            inline_request.inline_blocks.insert(inline_block_index, new_token)
            inline_block_index += 1
        while len(inline_request.inline_blocks) > inline_block_index:
            del inline_request.inline_blocks[inline_block_index]

    @staticmethod
    def __handle_next_special_character_remaining(
        inline_request: InlineRequest,
        inline_response: InlineResponse,
        remaining_line: str,
        current_string: str,
        current_string_unresolved: str,
    ) -> Tuple[str, str, str]:
        left_to_consume = inline_response.reduce_remaining_line_by
        if left_to_consume <= len(remaining_line):
            remaining_line = remaining_line[: -inline_response.reduce_remaining_line_by]
        else:
            left_to_consume -= len(remaining_line)
            remaining_line = ""
            if left_to_consume <= len(current_string):
                assert current_string_unresolved.endswith(
                    current_string[-left_to_consume:]
                ), "Unresolve string must end with part of the current string."
                current_string = current_string[:-left_to_consume]
                current_string_unresolved = current_string_unresolved[:-left_to_consume]
            else:
                left_to_consume -= len(current_string)
                current_string = ""
                current_string_unresolved = ""
                InlineTextBlockHelper.__handle_next_special_character_remaining_reduce(
                    inline_request, left_to_consume
                )
        return remaining_line, current_string, current_string_unresolved

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __handle_next_special_character(
        parser_properties: ParseBlockPassProperties,
        source_text: str,
        next_index: int,
        inline_request: InlineRequest,
        line_number: int,
        column_number: int,
        coalesced_stack: List[MarkdownToken],
        inline_response: InlineResponse,
        remaining_line: str,
        end_string: Optional[str],
        current_string: str,
        current_string_unresolved: str,
        inline_blocks: List[MarkdownToken],
        is_setext: bool,
        whitespace_to_recombine: Optional[str],
        para_owner: Optional[ParagraphMarkdownToken],
        tabified_text: Optional[str],
        tabified_remaining_line: Optional[str],
        pragma_line_numbers: Set[int],
    ) -> Tuple[
        InlineResponse,
        int,
        int,
        bool,
        bool,
        Optional[str],
        str,
        Optional[str],
        str,
        str,
        bool,
        Optional[str],
    ]:
        if InlineHandlerHelper.has_handler(source_text[next_index]):
            whitespace_to_add, was_new_line = None, False
            (
                inline_response,
                line_number,
                column_number,
                was_column_number_reset,
                did_line_number_change,
            ) = InlineHandlerHelper.process_inline_handled_character(
                parser_properties,
                source_text,
                next_index,
                inline_request,
                line_number,
                column_number,
                coalesced_stack,
                pragma_line_numbers,
            )
            if inline_response.reduce_remaining_line_by:
                (
                    remaining_line,
                    current_string,
                    current_string_unresolved,
                ) = InlineTextBlockHelper.__handle_next_special_character_remaining(
                    inline_request,
                    inline_response,
                    remaining_line,
                    current_string,
                    current_string_unresolved,
                )
        else:
            was_column_number_reset, did_line_number_change = False, False
            (
                whitespace_to_add,
                remaining_line,
                end_string,
                current_string,
                was_new_line,
                tabified_remaining_line,
            ) = InlineLineEndHelper.process_inline_new_line(
                source_text,
                next_index,
                inline_response,
                remaining_line,
                end_string,
                current_string,
                inline_blocks,
                is_setext,
                line_number,
                column_number,
                coalesced_stack,
                whitespace_to_recombine,
                para_owner,
                tabified_text,
                inline_request,
                tabified_remaining_line,
            )
        return (
            inline_response,
            line_number,
            column_number,
            was_column_number_reset,
            did_line_number_change,
            whitespace_to_add,
            remaining_line,
            end_string,
            current_string,
            current_string_unresolved,
            was_new_line,
            tabified_remaining_line,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __complete_inline_block_processing_build_token(
        current_string: str,
        end_string: Optional[str],
        starting_whitespace: str,
        is_setext: bool,
        inline_blocks: List[MarkdownToken],
        line_number: int,
        column_number: int,
    ) -> None:
        POGGER.debug("__cibp>current_string>$<", current_string)
        POGGER.debug("__cibp>starting_whitespace>$<", starting_whitespace)
        if (
            is_setext
            and end_string is None
            and (inline_blocks and inline_blocks[-1].is_inline_hard_break)
        ):
            new_index, ex_ws = ParserHelper.extract_spaces(current_string, 0)
            POGGER.debug("__cibp>new_index>$<", new_index)
            POGGER.debug("__cibp>b>$<", ex_ws)
            if new_index:
                end_string = f"{ex_ws}{ParserHelper.whitespace_split_character}"
                current_string = current_string[new_index:]
        POGGER.debug("__cibp>end_string>$<", end_string)
        inline_blocks.append(
            TextMarkdownToken(
                current_string,
                starting_whitespace,
                end_whitespace=end_string,
                line_number=line_number,
                column_number=column_number,
            )
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __create_new_text_token(
        inline_response: InlineResponse,
        current_string: str,
        inline_blocks: List[MarkdownToken],
        starting_whitespace: str,
        end_string: Optional[str],
        last_line_number: int,
        last_column_number: int,
        reset_current_string: bool,
    ) -> Tuple[bool, str, Optional[str]]:
        if inline_response.new_tokens:
            if current_string:
                inline_blocks.append(
                    TextMarkdownToken(
                        current_string,
                        starting_whitespace,
                        end_whitespace=end_string,
                        line_number=last_line_number,
                        column_number=last_column_number,
                    )
                )
                reset_current_string, starting_whitespace, end_string = (
                    True,
                    "",
                    None,
                )
            elif starting_whitespace:
                inline_blocks.append(
                    TextMarkdownToken(
                        "",
                        ParserHelper.create_replace_with_nothing_marker(
                            starting_whitespace
                        ),
                        line_number=last_line_number,
                        column_number=last_column_number,
                    )
                )
                starting_whitespace = ""

            inline_blocks.extend(inline_response.new_tokens)
        return reset_current_string, starting_whitespace, end_string

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __cleanup_after_handling(
        inline_response: InlineResponse,
        current_string: str,
        current_string_unresolved: str,
        remaining_line: str,
        tabified_remaining_line: Optional[str],
        reset_current_string: bool,
        end_string: Optional[str],
    ) -> Tuple[bool, str, Optional[str], str, str]:
        if inline_response.consume_rest_of_line:
            # POGGER.debug("consume_rest_of_line>>$<", remaining_line)
            (
                inline_response.new_string,
                inline_response.new_tokens,
                reset_current_string,
                remaining_line,
                end_string,
            ) = ("", [], True, "", None)
        else:
            # POGGER.debug("append_rest_of_line>>rem>>$<", remaining_line)
            proper_remaining_line = tabified_remaining_line or remaining_line
            current_string, current_string_unresolved = (
                InlineHelper.append_text(
                    current_string,
                    proper_remaining_line,
                ),
                InlineHelper.append_text(
                    current_string_unresolved, proper_remaining_line
                ),
            )
        return (
            reset_current_string,
            remaining_line,
            end_string,
            current_string,
            current_string_unresolved,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __handle_next_inline_character_finish_handling(
        line_number: int,
        column_number: int,
        split_para_space: Optional[List[str]],
        was_new_line: bool,
        coalesced_stack: List[MarkdownToken],
        remaining_line: str,
        did_line_number_change: bool,
        was_column_number_reset: bool,
        reset_current_string: bool,
        inline_blocks: List[MarkdownToken],
        old_inline_blocks_count: int,
        old_inline_blocks_last_token: Optional[MarkdownToken],
        source_text: str,
        whitespace_to_add: Optional[str],
        inline_response: InlineResponse,
        current_string: str,
        current_string_unresolved: str,
        last_line_number: int,
        last_column_number: int,
        end_string: Optional[str],
        pragma_line_numbers: Set[int],
    ) -> Tuple[
        int,
        int,
        Optional[List[str]],
        str,
        str,
        int,
        int,
        int,
        int,
        Optional[str],
    ]:
        (
            line_number,
            column_number,
            split_para_space,
        ) = InlineTextBlockHelper.__adjust_line_and_column_number(
            was_new_line,
            coalesced_stack,
            line_number,
            column_number,
            split_para_space,
            remaining_line,
            did_line_number_change,
            was_column_number_reset,
            pragma_line_numbers,
        )

        (
            current_string,
            current_string_unresolved,
            last_line_number,
            last_column_number,
        ) = InlineTextBlockHelper.__fix_variables_before_next_loop(
            reset_current_string,
            current_string,
            current_string_unresolved,
            inline_blocks,
            old_inline_blocks_count,
            old_inline_blocks_last_token,
            last_line_number,
            last_column_number,
            line_number,
            column_number,
        )

        assert (
            inline_response.new_index is not None
        ), "new_index must be defined by now."
        assert (
            inline_response.new_string is not None
        ), "new_string must be defined by now."
        (
            new_start_index,
            next_index,
            end_string,
            current_string,
            current_string_unresolved,
        ) = InlineTextBlockHelper.__complete_inline_loop(
            source_text,
            inline_response.new_index,
            end_string,
            whitespace_to_add,
            current_string,
            current_string_unresolved,
            inline_response.new_string_unresolved,
            inline_response.new_string,
            inline_response.original_string,
        )
        return (
            line_number,
            column_number,
            split_para_space,
            current_string,
            current_string_unresolved,
            last_line_number,
            last_column_number,
            new_start_index,
            next_index,
            end_string,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_line_and_column_number(
        was_new_line: bool,
        coalesced_stack: List[MarkdownToken],
        line_number: int,
        column_number: int,
        split_para_space: Optional[List[str]],
        remaining_line: str,
        did_line_number_change: bool,
        was_column_number_reset: bool,
        pragma_line_numbers: Set[int],
    ) -> Tuple[int, int, Optional[List[str]]]:
        if was_new_line:
            column_number = 1
            if coalesced_stack and coalesced_stack[-1].is_block_quote_start:
                block_quote_token = cast(BlockQuoteMarkdownToken, coalesced_stack[-1])
                POGGER.debug(
                    "coalesced_list[-1]..leading_text_index=$",
                    block_quote_token.leading_text_index,
                )
                assert (
                    block_quote_token.bleading_spaces is not None
                ), "Bleading spaces must be defined by now."
                split_leading_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                selected_split_length = len(
                    split_leading_spaces[block_quote_token.leading_text_index]
                )
                column_number += selected_split_length

            line_number += 1
            while line_number in pragma_line_numbers:
                line_number += 1
            assert (
                split_para_space
            ), "If we had a new line, we must have paragraph/setext text to handle."
            split_para_space = split_para_space[1:]
            column_number += len(split_para_space[0])
        elif not was_column_number_reset:
            column_number += len(remaining_line)
        else:
            assert did_line_number_change, "If here, the line number must have changed."
            if coalesced_stack and coalesced_stack[-1].is_block_quote_start:
                block_quote_token = cast(BlockQuoteMarkdownToken, coalesced_stack[-1])
                POGGER.debug(
                    "coalesced_list[-1].leading_text_index=$",
                    block_quote_token.leading_text_index,
                )
                assert (
                    block_quote_token.bleading_spaces is not None
                ), "Bleading spaces must be defined by now."
                split_leading_spaces = block_quote_token.bleading_spaces.split(
                    ParserHelper.newline_character
                )
                selected_split_length = len(
                    split_leading_spaces[block_quote_token.leading_text_index]
                )
                column_number += selected_split_length
        return line_number, column_number, split_para_space

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __fix_variables_before_next_loop(
        reset_current_string: bool,
        current_string: str,
        current_string_unresolved: str,
        inline_blocks: List[MarkdownToken],
        old_inline_blocks_count: int,
        old_inline_blocks_last_token: Optional[MarkdownToken],
        last_line_number: int,
        last_column_number: int,
        line_number: int,
        column_number: int,
    ) -> Tuple[str, str, int, int]:
        if reset_current_string:
            current_string, current_string_unresolved = "", ""

        if old_inline_blocks_count != len(inline_blocks) or (
            old_inline_blocks_last_token
            and old_inline_blocks_last_token != inline_blocks[-1]
        ):
            last_line_number, last_column_number = line_number, column_number
        return (
            current_string,
            current_string_unresolved,
            last_line_number,
            last_column_number,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    # pylint: disable=too-many-arguments
    def __complete_inline_loop(
        source_text: str,
        new_index: int,
        end_string: Optional[str],
        whitespace_to_add: Optional[str],
        current_string: str,
        current_string_unresolved: str,
        new_string_unresolved: Optional[str],
        new_string: str,
        original_string: Optional[str],
    ) -> Tuple[int, int, Optional[str], str, str]:
        POGGER.debug(
            "__complete_inline_loop--current_string>>$>>",
            current_string,
        )
        POGGER.debug(
            "__complete_inline_loop--new_string>>$>>",
            new_string,
        )
        POGGER.debug(
            "__complete_inline_loop--new_string_unresolved>>$>>",
            new_string_unresolved,
        )
        POGGER.debug(
            "__complete_inline_loop--original_string>>$>>",
            original_string,
        )

        POGGER.debug(
            "__complete_inline_loop--current_string>>$>>",
            current_string,
        )
        POGGER.debug(
            "__complete_inline_loop--end_string>>$>>",
            end_string,
        )
        POGGER.debug(
            "__complete_inline_loop--whitespace_to_add>>$>>",
            whitespace_to_add,
        )
        if original_string is not None:
            assert (
                new_string_unresolved is None
                or new_string_unresolved == original_string
            ), "new_string_unresolved must either be not defined or equal to the original_string."
            replaced_string = ParserHelper.create_replacement_markers(
                original_string, InlineHelper.append_text("", new_string)
            )
            current_string = f"{current_string}{replaced_string}"
        else:
            current_string = InlineHelper.append_text(current_string, new_string)
        POGGER.debug(
            "__complete_inline_loop--current_string>>$>>",
            current_string,
        )

        POGGER.debug(
            "new_string_unresolved>>$>>",
            new_string_unresolved,
        )
        if new_string == ParserHelper.newline_character and end_string:
            split_end_string = end_string.split(ParserHelper.newline_character)
            POGGER.debug(
                "split_end_string>>$>>",
                split_end_string,
            )
            assert (
                len(split_end_string) >= 2
            ), "end_string must be split into at least 2 parts"
            new_string = split_end_string[len(split_end_string) - 2] + new_string

        current_string_unresolved = (
            f"{current_string_unresolved}{new_string_unresolved}"
            if new_string_unresolved
            else InlineHelper.append_text(current_string_unresolved, new_string)
        )

        POGGER.debug(
            "__complete_inline_loop--current_string_unresolved>>$>>",
            current_string_unresolved,
        )

        start_index = new_index
        next_index = ParserHelper.index_any_of(
            source_text,
            InlineHandlerHelper.valid_inline_text_block_sequence_starts,
            start_index,
        )
        return (
            start_index,
            next_index,
            end_string,
            current_string,
            current_string_unresolved,
        )

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to implement a plugin that looks for inconsistent styles for thematic breaks.
"""

from typing import List, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.thematic_break_markdown_token import ThematicBreakMarkdownToken


class RuleMd035(RulePlugin):
    """
    Class to implement a plugin that looks for inconsistent styles for thematic breaks.
    """

    __consistent_style = "consistent"

    def __init__(self) -> None:
        super().__init__()
        self.__rule_style: str = ""
        self.__actual_style: str = ""

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="hr-style",
            plugin_id="MD035",
            plugin_enabled_by_default=True,
            plugin_description="Horizontal rule style",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md035.md",
            plugin_configuration="style",
            plugin_supports_fix=True,
        )

    @classmethod
    def __validate_configuration_style(cls, found_value: str) -> None:
        if found_value == RuleMd035.__consistent_style:
            return
        if found_value != found_value.strip(" "):
            raise ValueError(
                "Allowable values cannot including leading or trailing spaces."
            )
        if is_valid := bool(found_value):
            valid_character = None
            for next_character in found_value:
                if next_character not in " _-*":
                    is_valid = False
                    break
                if next_character != " ":
                    if valid_character is None:
                        valid_character = next_character
                    elif next_character != valid_character:
                        is_valid = False
                        break
        if not is_valid:
            raise ValueError(
                f"Allowable values are: {RuleMd035.__consistent_style}, "
                + "'---', '***', `___`, or any other horizontal rule text."
            )

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__rule_style = self.plugin_configuration.get_string_property_with_default(
            "style",
            RuleMd035.__consistent_style,
            valid_value_fn=self.__validate_configuration_style,
        )
        if self.__rule_style != RuleMd035.__consistent_style:
            self.__actual_style = self.__rule_style

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("style", self.__rule_style),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        if self.__rule_style == RuleMd035.__consistent_style:
            self.__actual_style = ""

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if not token.is_thematic_break:
            return
        break_token = cast(ThematicBreakMarkdownToken, token)
        if self.__actual_style:
            if self.__actual_style != break_token.rest_of_line:
                if context.in_fix_mode:
                    self.register_fix_token_request(
                        context,
                        token,
                        "next_token",
                        "start_character",
                        self.__actual_style[0],
                    )
                    self.register_fix_token_request(
                        context,
                        token,
                        "next_token",
                        "rest_of_line",
                        self.__actual_style,
                    )
                else:
                    extra_data = f"Expected: {self.__actual_style}, Actual: {break_token.rest_of_line}"
                    self.report_next_token_error(
                        context, token, extra_error_information=extra_data
                    )
        else:
            self.__actual_style = break_token.rest_of_line



"""
Module to provide detection of the list starts.
"""

import logging
import string
from typing import Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.thematic_leaf_block_processor import (
    ThematicLeafBlockProcessor,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.stack_token import ListStackToken

POGGER = ParserLogger(logging.getLogger(__name__))


class ListBlockStartsHelper:
    """
    Class to provide detection of the list starts.
    """

    __ulist_start_characters = "-+*"
    __olist_start_characters = ".)"

    # pylint: disable=too-many-arguments
    @staticmethod
    def is_ulist_start(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        skip_whitespace_check: bool,
        adj_ws: Optional[str],
    ) -> Tuple[bool, int, Optional[int], Optional[int]]:
        """
        Determine if we have the start of an un-numbered list.
        """
        POGGER.debug("is_ulist_start>>pre>>")
        POGGER.debug("is_ulist_start>>start_index>>$<<", start_index)
        POGGER.debug("is_ulist_start>>adj_ws>>$<<", adj_ws)
        POGGER.debug("is_ulist_start>>extracted_whitespace>>$<<", extracted_whitespace)
        ex_ws = extracted_whitespace if adj_ws is None else adj_ws
        (
            check_ws,
            parent_indent,
        ) = ListBlockStartsHelper.__adjust_whitespace_for_nested_lists(
            parser_state,
            ex_ws,
            line_to_parse,
            start_index,
        )
        POGGER.debug("skip_whitespace_check>>$", skip_whitespace_check)
        POGGER.debug("parent_indent>>$", parent_indent)
        POGGER.debug("len(check_ws)>>$", len(check_ws))

        if (
            TabHelper.is_length_less_than_or_equal_to(check_ws, 3 + parent_indent)
            or skip_whitespace_check
        ):
            adj_extracted_whitespace = (
                extracted_whitespace[parent_indent:]
                if parent_indent
                else extracted_whitespace
            )
            is_start = ListBlockStartsHelper.__is_start_ulist(
                line_to_parse, start_index, adj_extracted_whitespace
            )
        else:
            is_start = False
        if is_start:
            (
                is_start,
                after_all_whitespace_index,
            ) = ListBlockStartsHelper.__is_start_phase_one(
                parser_state, line_to_parse, start_index, False
            )
        else:
            after_all_whitespace_index = -1
        if is_start:
            is_start = ListBlockStartsHelper.__is_start_phase_two(
                parser_state,
                line_to_parse[start_index],
                True,
                False,
                after_all_whitespace_index,
                line_to_parse,
                start_index,
            )

        return is_start, after_all_whitespace_index, start_index, 0
        # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def is_olist_start(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        skip_whitespace_check: bool,
        adj_ws: Optional[str] = None,
    ) -> Tuple[bool, int, Optional[int], Optional[int]]:
        """
        Determine if we have the start of a numbered or ordered list.
        """

        POGGER.debug("is_olist_start>>pre>>")
        POGGER.debug("is_olist_start>>start_index>>$<<", start_index)
        POGGER.debug("is_olist_start>>adj_ws>>$<<", adj_ws)
        POGGER.debug("is_olist_start>>extracted_whitespace>>$<<", extracted_whitespace)

        ex_ws = extracted_whitespace if adj_ws is None else adj_ws
        (
            check_ws,
            parent_indent,
        ) = ListBlockStartsHelper.__adjust_whitespace_for_nested_lists(
            parser_state,
            ex_ws,
            line_to_parse,
            start_index,
        )
        POGGER.debug("after_adjust>>ws=$=", check_ws)
        POGGER.debug("after_adjust>>parent_indent=$=", parent_indent)

        POGGER.debug("skip_whitespace_check>>$", skip_whitespace_check)
        POGGER.debug("len(check_ws)>>$", len(check_ws))

        if (
            TabHelper.is_length_less_than_or_equal_to(check_ws, 3 + parent_indent)
            or skip_whitespace_check
        ):
            (
                is_start,
                index,
                number_of_digits,
                is_not_one,
            ) = ListBlockStartsHelper.__is_start_olist(line_to_parse, start_index)
        else:
            is_start, index, number_of_digits, is_not_one = False, None, None, False
        if is_start:
            assert (
                index is not None and is_not_one is not None
            ), "If is_start, these must be valid."
            (
                is_start,
                after_all_whitespace_index,
            ) = ListBlockStartsHelper.__is_start_phase_one(
                parser_state, line_to_parse, index, is_not_one
            )
            if is_start:
                is_start = ListBlockStartsHelper.__is_start_phase_two(
                    parser_state,
                    line_to_parse[index],
                    False,
                    is_not_one,
                    after_all_whitespace_index,
                    line_to_parse,
                    start_index,
                )
        else:
            after_all_whitespace_index = -1

        return is_start, after_all_whitespace_index, index, number_of_digits

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_whitespace_for_nested_lists(
        parser_state: ParserState,
        adj_ws: str,
        line_to_parse: str,
        start_index: int,
    ) -> Tuple[str, int]:
        (
            child_list_token,
            parent_list_token,
        ) = ListBlockStartsHelper.__determine_child_and_parent_tokens(parser_state)
        POGGER.debug("len(adj_ws)>>$", len(adj_ws))

        if child_list_token and parent_list_token:
            parent_indent, child_indent = (
                parent_list_token.indent_level,
                child_list_token.indent_level,
            )
            POGGER.debug("parent_indent>>$", parent_indent)
            POGGER.debug("child_indent>>$", child_indent)
            if len(adj_ws) > parent_indent and len(adj_ws) < child_indent:
                adj_ws = adj_ws[parent_indent:]
        elif child_list_token:
            POGGER.debug("current_start>>$", child_list_token.matching_markdown_token)
            POGGER.debug(
                "current_start.last_new_list_token>>$",
                child_list_token.last_new_list_token,
            )
            POGGER.debug("line_to_parse>>:$:", line_to_parse)
            POGGER.debug("start_index>>$", start_index)

            indent_level = (
                child_list_token.last_new_list_token.indent_level
                if child_list_token.last_new_list_token
                else child_list_token.indent_level
            )
            parent_indent = (
                child_list_token.indent_level if start_index >= indent_level else 0
            )
        else:
            parent_indent = 0
        return adj_ws, parent_indent

    @staticmethod
    def __is_start_phase_one(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        is_not_one: bool,
    ) -> Tuple[bool, int]:
        start_index += 1
        line_to_parse_size = len(line_to_parse)
        after_all_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            line_to_parse, start_index
        )
        POGGER.debug(
            "after_all_whitespace_index>>$>>len>>$",
            after_all_whitespace_index,
            line_to_parse_size,
        )
        at_end_of_line = after_all_whitespace_index == line_to_parse_size
        POGGER.debug("at_end_of_line>>$", at_end_of_line)

        is_in_paragraph = parser_state.token_stack[-1].is_paragraph
        is_paragraph_in_list = (
            parser_state.token_stack[-2].is_list if is_in_paragraph else False
        )

        is_block_within_list = False
        if (
            parser_state.token_stack[-1].is_fenced_code_block
            or parser_state.token_stack[-1].is_html_block
        ) and parser_state.token_stack[-2].is_list:
            POGGER.debug("start_index>>$", start_index)
            matching_list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[-2].matching_markdown_token,
            )
            POGGER.debug("matching_list_token>>$", matching_list_token)
            POGGER.debug(
                "matching_list_token.indent_level>>$", matching_list_token.indent_level
            )
            is_block_within_list = start_index > matching_list_token.indent_level
            POGGER.debug("is_block_within_list>>$", is_block_within_list)

        is_paragraph_continuation = (
            is_in_paragraph
            and not is_paragraph_in_list
            and (at_end_of_line or is_not_one)
        )
        is_start = (
            not is_paragraph_continuation
            and not is_block_within_list
            and (
                ParserHelper.is_character_at_index_whitespace(
                    line_to_parse, start_index
                )
                or ((start_index) == line_to_parse_size)
            )
        )
        return is_start, after_all_whitespace_index

    # pylint: disable=too-many-arguments
    @staticmethod
    def __is_start_phase_two(
        parser_state: ParserState,
        xx_seq: str,
        is_unordered_list: bool,
        is_not_one: bool,
        after_all_whitespace_index: int,
        line_to_parse: str,
        start_index: int,
    ) -> bool:
        (is_in_paragraph, at_end_of_line) = (
            parser_state.token_stack[-1].is_paragraph,
            (after_all_whitespace_index == len(line_to_parse)),
        )

        if is_in_paragraph:
            (
                is_first_item_in_list,
                is_sub_list,
            ) = ListBlockStartsHelper.__calculate_starts_within_paragraph(
                parser_state, line_to_parse, start_index, is_unordered_list, xx_seq
            )
        else:
            is_first_item_in_list, is_sub_list = False, False

        POGGER.debug(
            "is_in_para>>$(>>EOL>$>>is_not_one>$)>>is_first>$>>is_sub_list>$",
            is_in_paragraph,
            at_end_of_line,
            is_not_one,
            is_first_item_in_list,
            is_sub_list,
        )
        return not (
            is_in_paragraph
            and (at_end_of_line or is_not_one)
            and is_first_item_in_list
            and is_sub_list
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __is_start_ulist(
        line_to_parse: str, start_index: int, extracted_whitespace: str
    ) -> bool:
        # Thematic breaks have precedence, so stop a list start if we find one.
        if is_start := ParserHelper.is_character_at_index_one_of(
            line_to_parse, start_index, ListBlockStartsHelper.__ulist_start_characters
        ):
            assert extracted_whitespace is not None
            is_break, _ = ThematicLeafBlockProcessor.is_thematic_break(
                line_to_parse, start_index, extracted_whitespace
            )
            is_start = is_start and not is_break
        return is_start

    @staticmethod
    def __is_start_olist(
        line_to_parse: str, start_index: int
    ) -> Tuple[bool, Optional[int], Optional[int], Optional[bool]]:
        if is_start := ParserHelper.is_character_at_index_one_of(
            line_to_parse, start_index, string.digits
        ):
            index, olist_index_number = (
                ParserHelper.collect_while_one_of_characters_verified(
                    line_to_parse, start_index, string.digits
                )
            )
            number_of_digits = len(olist_index_number)

            POGGER.debug("olist?$<<count>>$<<", olist_index_number, number_of_digits)
            is_not_one = olist_index_number != "1"
            is_start = (
                number_of_digits <= 9
                and ParserHelper.is_character_at_index_one_of(
                    line_to_parse, index, ListBlockStartsHelper.__olist_start_characters
                )
            )
        else:
            index, number_of_digits, is_not_one = None, None, None

        POGGER.debug("is_olist_start>>$", is_start)
        return is_start, index, number_of_digits, is_not_one

    @staticmethod
    def __determine_child_and_parent_tokens(
        parser_state: ParserState,
    ) -> Tuple[Optional[ListStackToken], Optional[ListStackToken]]:
        child_list_token, parent_list_token = None, None
        if parser_state.token_stack[-1].is_list:
            child_list_token = cast(ListStackToken, parser_state.token_stack[-1])
            if (
                len(parser_state.token_stack) > 1
                and parser_state.token_stack[-2].is_list
            ):
                parent_list_token = cast(ListStackToken, parser_state.token_stack[-2])
        elif len(parser_state.token_stack) > 1 and parser_state.token_stack[-2].is_list:
            child_list_token = cast(ListStackToken, parser_state.token_stack[-2])
            if (
                len(parser_state.token_stack) > 2
                and parser_state.token_stack[-3].is_list
            ):
                parent_list_token = cast(ListStackToken, parser_state.token_stack[-3])
        POGGER.debug("child_list_token>>$", child_list_token)
        POGGER.debug("parent_list_token>>$", parent_list_token)
        return child_list_token, parent_list_token

    @staticmethod
    def __calculate_starts_within_paragraph(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        is_unordered_list: bool,
        xx_seq: str,
    ) -> Tuple[bool, bool]:
        is_first_item_in_list = True
        if not parser_state.token_stack[-2].is_list:
            POGGER.debug("top of stack is not list>>$", parser_state.token_stack[-2])
        else:
            list_stack_token = cast(ListStackToken, parser_state.token_stack[-2])
            if is_unordered_list and list_stack_token.is_ordered_list:
                POGGER.debug("top of stack is ordered list>>$", list_stack_token)
            elif xx_seq != list_stack_token.list_character[-1]:
                POGGER.debug(
                    "xx>>$!=$",
                    line_to_parse[start_index],
                    list_stack_token.list_character,
                )
            else:
                is_first_item_in_list = start_index >= list_stack_token.indent_level
                POGGER.debug(
                    "start_index>>$>=$",
                    start_index,
                    list_stack_token.indent_level,
                )
        POGGER.debug("is_first_item_in_list>>$", is_first_item_in_list)

        if parser_state.token_stack[-2].is_list:
            list_token = cast(ListStackToken, parser_state.token_stack[-2])
            is_sub_list = start_index >= list_token.indent_level
        else:
            is_sub_list = False

        return is_first_item_in_list, is_sub_list



"""
Module to implement a plugin that looks for text in a paragraph where a line starts
with what could be a closed atx heading, except there is no spaces between the hashes
and the text of the heading, either at the start, end, or both.
"""

import re
from typing import Any, Optional, cast

from pymarkdown.plugin_manager.plugin_details import PluginDetails
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.plugins.rule_md_018 import StartOfLineTokenParser
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken


class MyStartOfLineTokenParser(StartOfLineTokenParser):
    """
    Local implementation of the token parser.
    """

    def __init__(self, owner: "RuleMd020") -> None:
        super().__init__()
        self.__owner = owner

    # pylint: disable=too-many-arguments
    def check_start_of_line(
        self,
        combined_text: str,
        context: Any,
        token: MarkdownToken,
        line_number_delta: int,
        column_number_delta: int,
    ) -> None:
        """
        Check for a pattern at the start of the line.
        """
        if re.search(r"^[ ]{0,3}#{1,6}.*#+[ ]*$", combined_text):
            self.__owner.report_next_token_error(
                context,
                token,
                line_number_delta=line_number_delta
                + context.calc_pragma_offset(token, line_number_delta),
                column_number_delta=column_number_delta,
            )

    # pylint: enable=too-many-arguments


class RuleMd020(RulePlugin):
    """
    Class to implement a plugin that looks for text in a paragraph where a line starts
    with what could be a closed atx heading, except there is no spaces between the hashes
    and the text of the heading, either at the start, end, or both.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__token_parser = MyStartOfLineTokenParser(self)
        self.__is_in_normal_atx = False
        self.__last_atx_token: Optional[MarkdownToken] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetails(
            plugin_name="no-missing-space-closed-atx",
            plugin_id="MD020",
            plugin_enabled_by_default=True,
            plugin_description="No space present inside of the hashes on a possible Atx Closed Heading.",
            plugin_version="0.5.1",
            plugin_interface_version=1,
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md020.md",
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__token_parser.starting_new_file()
        self.__is_in_normal_atx = False
        self.__last_atx_token = None

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        self.__token_parser.next_token(context, token)
        if not token.is_atx_heading_end and self.__is_in_normal_atx:
            self.__last_atx_token = token

        if token.is_atx_heading:
            self.__is_in_normal_atx = True
        elif token.is_atx_heading_end:
            assert self.__last_atx_token is not None
            if self.__is_in_normal_atx and self.__last_atx_token.is_text:
                text_token = cast(TextMarkdownToken, self.__last_atx_token)
                if text_token.token_text.endswith(
                    "#"
                ) and not text_token.token_text.endswith("\\\b#"):
                    regex_match = re.search(r"\#+$", text_token.token_text)
                    assert regex_match is not None
                    self.report_next_token_error(
                        context,
                        self.__last_atx_token,
                        column_number_delta=regex_match.start(),
                    )
            self.__is_in_normal_atx = False



"""
Module to help with the parsing of autolink inline elements.
"""

import logging
import re
import string
from typing import Optional, Tuple, Union, cast

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.html.html_raw_helper import HtmlRawHelper
from pymarkdown.inline.inline_request import InlineRequest
from pymarkdown.inline.inline_response import InlineResponse
from pymarkdown.tokens.email_autolink_markdown_token import EmailAutolinkMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.raw_html_markdown_token import RawHtmlMarkdownToken
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken
from pymarkdown.tokens.uri_autolink_markdown_token import UriAutolinkMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))


class InlineAutoLinkHelper:
    __scheme_end_character = ":"
    __valid_scheme_characters = f"{string.ascii_letters}{string.digits}.-+"
    __valid_email_regex = (
        "^[a-zA-Z0-9.!#$%&'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}"
        + "[a-zA-Z0-9])?(?:\\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$"
    )

    angle_bracket_start = "<"
    __angle_bracket_end = ">"

    """
    Class to help with the parsing of autolink inline elements.
    """

    @staticmethod
    def __handle_raw_html(
        between_brackets: str,
        remaining_line: str,
        inline_request: InlineRequest,
        new_column_number: int,
        closing_angle_index: int,
    ) -> Tuple[Optional[Union[RawHtmlMarkdownToken, TextMarkdownToken]], str, int]:
        assert (
            inline_request.line_number is not None
        ), "Request line number cannot be None."
        new_token, after_index = HtmlRawHelper.parse_raw_html(
            between_brackets,
            remaining_line,
            inline_request.line_number,
            new_column_number,
            inline_request,
        )
        if after_index != -1:
            assert (
                new_token is not None
            ), "If after_index is valid, new_token must also be valid."
            closing_angle_index = after_index + inline_request.next_index + 1
            if new_token.is_inline_raw_html:
                html_token = cast(RawHtmlMarkdownToken, new_token)
                between_brackets = html_token.raw_tag
            else:
                between_brackets = between_brackets
        return new_token, between_brackets, closing_angle_index

    @staticmethod
    def handle_angle_brackets(
        parser_properties: ParseBlockPassProperties, inline_request: InlineRequest
    ) -> InlineResponse:
        """
        Given an open angle bracket, determine which of the three possibilities it is.
        """
        _ = parser_properties
        closing_angle_index = inline_request.source_text.find(
            InlineAutoLinkHelper.__angle_bracket_end, inline_request.next_index
        )
        if closing_angle_index not in (-1, inline_request.next_index + 1):
            between_brackets, remaining_line = (
                inline_request.source_text[
                    inline_request.next_index + 1 : closing_angle_index
                ],
                inline_request.source_text[inline_request.next_index + 1 :],
            )
            closing_angle_index += 1

            assert (
                inline_request.line_number is not None
                and inline_request.column_number is not None
                and inline_request.remaining_line is not None
            ), "Proper processing must leave these in a valid state."
            new_column_number = inline_request.column_number + len(
                inline_request.remaining_line
            )

            new_token: Optional[MarkdownToken] = (
                InlineAutoLinkHelper.__parse_valid_uri_autolink(
                    between_brackets, inline_request.line_number, new_column_number
                )
            )
            if not new_token:
                new_token = InlineAutoLinkHelper.__parse_valid_email_autolink(
                    between_brackets, inline_request.line_number, new_column_number
                )
            if not new_token:
                (
                    new_token,
                    between_brackets,
                    closing_angle_index,
                ) = InlineAutoLinkHelper.__handle_raw_html(
                    between_brackets,
                    remaining_line,
                    inline_request,
                    new_column_number,
                    closing_angle_index,
                )
        else:
            new_token, between_brackets = None, None

        inline_response = InlineResponse()
        if new_token:
            (
                inline_response.new_string,
                inline_response.new_index,
                inline_response.new_tokens,
                between_brackets,
            ) = (
                "",
                closing_angle_index,
                [new_token],
                f"{InlineAutoLinkHelper.angle_bracket_start}{between_brackets}{InlineAutoLinkHelper.__angle_bracket_end}",
            )
        else:
            inline_response.new_string, inline_response.new_index, between_brackets = (
                InlineAutoLinkHelper.angle_bracket_start,
                inline_request.next_index + 1,
                InlineAutoLinkHelper.angle_bracket_start,
            )

        (
            inline_response.delta_line_number,
            inline_response.delta_column_number,
        ) = ParserHelper.calculate_deltas(between_brackets)
        return inline_response

    @staticmethod
    def __parse_valid_email_autolink(
        text_to_parse: str, line_number: int, column_number: int
    ) -> Optional[EmailAutolinkMarkdownToken]:
        """
        Parse a possible email autolink and determine if it is valid.
        """
        return (
            EmailAutolinkMarkdownToken(text_to_parse, line_number, column_number)
            if re.match(InlineAutoLinkHelper.__valid_email_regex, text_to_parse)
            else None
        )

    @staticmethod
    def __parse_valid_uri_autolink(
        text_to_parse: str, line_number: int, column_number: int
    ) -> Optional[UriAutolinkMarkdownToken]:
        """
        Parse a possible uri autolink and determine if it is valid.
        """

        if (
            InlineAutoLinkHelper.angle_bracket_start not in text_to_parse
            and text_to_parse[0] in string.ascii_letters
        ):
            path_index, uri_scheme = (
                ParserHelper.collect_while_one_of_characters_verified(
                    text_to_parse, 1, InlineAutoLinkHelper.__valid_scheme_characters
                )
            )
            uri_scheme, text_to_parse_size = f"{text_to_parse[0]}{uri_scheme}", len(
                text_to_parse
            )
            if (
                2 <= len(uri_scheme) <= 32
                and path_index < text_to_parse_size
                and text_to_parse[path_index]
                == InlineAutoLinkHelper.__scheme_end_character
            ):
                path_index += 1
                while (
                    path_index < text_to_parse_size
                    and ord(text_to_parse[path_index]) > 32
                ):
                    path_index += 1
                if path_index == text_to_parse_size:
                    return UriAutolinkMarkdownToken(
                        text_to_parse, line_number, column_number
                    )
        else:
            uri_scheme, path_index = "", -1
        return None
